/******************************************************************************/
/*                                                                            */
/*  Cedocida DV-Codec                                                         */
/*                                                                            */
/*  Copyright (C) 2005 Andreas Dittrich <dittrich@cithraidt.de>               */
/*  Project start: 15.12.2003                                                 */
/*  Last change  : 15.06.2007                                                 */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/*  This program is free software; you can redistribute it and/or modify      */
/*  it under the terms of the GNU General Public License as published by      */
/*  the Free Software Foundation; either version 2 of the License, or         */
/*  (at your option) any later version.                                       */
/*                                                                            */
/*  This program is distributed in the hope that it will be useful,           */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             */
/*  GNU General Public License for more details.                              */
/*                                                                            */
/*  You should have received a copy of the GNU General Public License         */
/*  along with this program; if not, write to the Free Software               */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
/*                                                                            */
/******************************************************************************/


#include "codec_vfw.h"
#include <stdio.h>
#include <string.h>

#include "inttypes.h"
#include "config.h"
#include "encode.h"
#include "decode.h"

#include "resource.h"


#define ABS(x) ((x) < 0 ? (-(x)) : (x))


typedef struct {
	void *henc;
	void *hdec;
	dvformat_t dvfmt;
	configvars_t config;
} driver_context_t;


HMODULE hmodule_cedocida = NULL;


__declspec(dllexport) LRESULT WINAPI DriverProc(
	DWORD_PTR dwDriverID, 
	HDRVR hDriver, 
	UINT uiMessage, 
	LPARAM lParam1, 
	LPARAM lParam2) 
{
	void * hnd = (void*)dwDriverID; 

	switch (uiMessage) {

		/****************************************
			standard driver messages
		****************************************/
		case DRV_INSTALL: return DRV_OK;
		case DRV_REMOVE: return DRV_OK;
		case DRV_ENABLE: return DRV_OK;
		case DRV_DISABLE: return 0;
		case DRV_LOAD: return DRV_OK;
		case DRV_OPEN: return my_drv_open(lParam2);
		case DRV_CLOSE: return my_drv_close(hnd);
		case DRV_FREE: return DRV_OK;
		case DRV_QUERYCONFIGURE: return DRV_OK;
		case DRV_CONFIGURE: return my_drv_configure(hnd, lParam1);

		/****************************************
			state messages
		****************************************/
		case ICM_ABOUT: return ICERR_UNSUPPORTED;
		case ICM_CONFIGURE: return my_icm_configure(hnd, lParam1);
		case ICM_GETSTATE: return my_icm_getstate(hnd, lParam1, lParam2);
		case ICM_SETSTATE: return my_icm_setstate(hnd, lParam1, lParam2);
		case ICM_GET: return 0;
		case ICM_GETQUALITY:
		case ICM_SETQUALITY:
		case ICM_GETDEFAULTQUALITY:
		case ICM_GETDEFAULTKEYFRAMERATE: return ICERR_UNSUPPORTED;
		case ICM_GETINFO: return my_icm_getinfo(hnd, lParam1, lParam2);

		/****************************************
			compress messages
		****************************************/
		case ICM_COMPRESS_QUERY: return my_icm_compress_query(hnd, lParam1, lParam2);
		case ICM_COMPRESS_GET_FORMAT: return my_icm_compress_get_format(hnd, lParam1, lParam2);
		case ICM_COMPRESS_GET_SIZE: return my_icm_get_size(hnd, lParam1);
		case ICM_COMPRESS_BEGIN: return my_icm_compress_begin(hnd, lParam1, lParam2);
		case ICM_COMPRESS: return my_icm_compress(hnd, lParam1, lParam2);
		case ICM_COMPRESS_END: return my_icm_compress_end(hnd);

		/****************************************
			decompress messages
		****************************************/
		case ICM_DECOMPRESS_GET_PALETTE: return ICERR_BADFORMAT;
		case ICM_DECOMPRESS_SET_PALETTE: return ICERR_UNSUPPORTED;
		case ICM_DECOMPRESS_QUERY: return my_icm_decompress_query(hnd, lParam1, lParam2);
		case ICM_DECOMPRESS_GET_FORMAT: return my_icm_decompress_get_format(hnd, lParam1, lParam2);
		case ICM_DECOMPRESS_BEGIN: return my_icm_decompress_begin(hnd, lParam1, lParam2);
		case ICM_DECOMPRESS: return my_icm_decompress(hnd, lParam1, lParam2);
		case ICM_DECOMPRESS_END: return my_icm_decompress_end(hnd);
	}

	if (uiMessage < DRV_USER)
		return DefDriverProc(dwDriverID, hDriver, uiMessage, lParam1, lParam2);

	return ICERR_UNSUPPORTED;
}


BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID) 
{
	if (fdwReason == DLL_PROCESS_ATTACH)
	{
		hmodule_cedocida = (HMODULE)hinstDLL;
	}
	
	return TRUE;
}


/****************************************
	helper functions
****************************************/

static int save_configuration(configvars_t const *pconfig, dvformat_t dvfmt);
static int load_configuration(configvars_t *pconfig, dvformat_t dvfmt);
static int default_configuration(configvars_t *pconfig);


static uint8_t lower(uint8_t ch)
{
	if (ch >= (uint8_t)'A' && ch <= (uint8_t)'Z')
		return ch - (uint8_t)'A' + (uint8_t)'a';

	return ch;
}


static int is_equal_fourcc(DWORD fcc1, DWORD fcc2)
{
	return 
		lower((fcc1 >> 24) & 0xFF) == lower((fcc2 >> 24) & 0xFF) &&
		lower((fcc1 >> 16) & 0xFF) == lower((fcc2 >> 16) & 0xFF) &&
		lower((fcc1 >>  8) & 0xFF) == lower((fcc2 >>  8) & 0xFF) &&
		lower((fcc1      ) & 0xFF) == lower((fcc2      ) & 0xFF);
}


static int get_colorformat(BITMAPINFOHEADER* pbihInp)
{
	if (!pbihInp) 
		return 0;

	int fourcc    = pbihInp->biCompression;
	int bitcount  = pbihInp->biBitCount;

	if ((fourcc == 0 || fourcc == ' BID') && (bitcount == 32))  return COLORFORMAT_RGB32; else
	if ((fourcc == 0 || fourcc == ' BID') && (bitcount == 24))  return COLORFORMAT_RGB24; else
	if (fourcc == FOURCC_YUY2)                                  return COLORFORMAT_YUY2; else
	if (fourcc == FOURCC_YV12)                                  return COLORFORMAT_YV12; else
		return COLORFORMAT_UNKNOWN;
}


static int get_input_colorformat(configvars_t *pconfig, BITMAPINFOHEADER* pbihInp)
{
	int color_format = get_colorformat(pbihInp);

	if (color_format < 0)
		return color_format;

	color_format &= pconfig->encoder_input_format;

	if (color_format == 0)
		return COLORFORMAT_NA;

	return color_format;
}


static int get_output_colorformat(configvars_t *pconfig, BITMAPINFOHEADER* pbihInp)
{
	int color_format = get_colorformat(pbihInp);

	if (color_format < 0)
		return color_format;

	color_format &= pconfig->decoder_output_format;

	if (color_format == 0)
		return COLORFORMAT_NA;

	return color_format;
}


static DWORD dvfmt_to_fourcc(dvformat_t dvfmt)
{
	return
		(dvfmt == DVFORMAT_DV25) ? FOURCC_dv25 :
		(dvfmt == DVFORMAT_DV50) ? FOURCC_dv50 :
		FOURCC_dvsd;
}


/****************************************
	standard driver messages
****************************************/

static INT_PTR CALLBACK ConfigureDialogProc(HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);

LONG my_drv_configure(void *hnd, LPARAM lParam1)
{
	driver_context_t * const hdrv = (driver_context_t*)hnd;
	if (hdrv == NULL)
		return DRVCNF_CANCEL;

	HWND hwnd = (HWND)lParam1;

	configvars_t config = hdrv->config;

	INT_PTR res = DialogBoxParam(
		hmodule_cedocida, 
		MAKEINTRESOURCE(IDD_DIALOG_CONFIGURE), 
		hwnd, 
		ConfigureDialogProc, 
		(LPARAM)&config);

	if (res == 1)	// 0 corresponds to "cancel", 1 to "ok"
	{
		save_configuration(&config, hdrv->dvfmt);
		hdrv->config = config;
	}

	return DRV_OK;
}


DWORD_PTR my_drv_open(LPARAM lParam2) 
{
	ICOPEN * icinfo = (ICOPEN*)lParam2;

	dvformat_t dvfmt = DVFORMAT_DVSD;

	if (icinfo)
	{
		if (icinfo->fccType != ICTYPE_VIDEO) 
		{
			icinfo->dwError = ICERR_UNSUPPORTED;
			return NULL;
		}

		if (is_equal_fourcc(icinfo->fccHandler, FOURCC_dvsd)) dvfmt = DVFORMAT_DVSD; else
		if (is_equal_fourcc(icinfo->fccHandler, FOURCC_dv25)) dvfmt = DVFORMAT_DV25; else
		if (is_equal_fourcc(icinfo->fccHandler, FOURCC_dv50)) dvfmt = DVFORMAT_DV50; else
		{
			icinfo->dwError = ICERR_UNSUPPORTED;
			return NULL;
		}

		icinfo->dwError = ICERR_OK;
	}

	driver_context_t * hdrv = (driver_context_t*)malloc(sizeof(driver_context_t));
	if (hdrv == NULL)
		return NULL;

	memset(hdrv, 0x00, sizeof(driver_context_t));

	hdrv->dvfmt = dvfmt;

	load_configuration(&hdrv->config, hdrv->dvfmt); // this shall be the only place, where the configuration is loaded from the registry

	return (DWORD_PTR)hdrv;
}


LONG my_drv_close(void *hnd) 
{
	driver_context_t * hdrv = (driver_context_t*)hnd;

	if (hdrv)
	{
		if (hdrv->henc)
			encoder_close(hdrv->henc);

		if (hdrv->hdec)
			decoder_close(hdrv->hdec);

		free(hdrv);
	}

	return (LRESULT)1L;
}


/****************************************
	state messages
****************************************/

LONG my_icm_configure(void *hnd, LPARAM lParam1)
{
	driver_context_t * const hdrv = (driver_context_t*)hnd;
	if (hdrv == NULL)
		return ICERR_BADHANDLE;

	if (lParam1 == -1)
		return ICERR_OK;

	my_drv_configure(hdrv, lParam1);

	return ICERR_OK;
}


LONG my_icm_getstate(void *hnd, LPARAM lParam1, LPARAM lParam2)
{
	driver_context_t * const hdrv = (driver_context_t*)hnd;
	if (hdrv == NULL)
		return ICERR_BADHANDLE;

	configvars_t * pconfig = (configvars_t*)(void*)lParam1;

	int nbytes = (int)lParam2;

	if (pconfig == NULL)
		return sizeof(configvars_t);

	if (nbytes != sizeof(configvars_t))
		return ICERR_BADPARAM;

	*pconfig = hdrv->config;
	
	return ICERR_OK;
}


LONG my_icm_setstate(void *hnd, LPARAM lParam1, LPARAM lParam2)
{
	driver_context_t * const hdrv = (driver_context_t*)hnd;
	if (hdrv == NULL)
		return ICERR_BADHANDLE;

	configvars_t const * pconfig = (configvars_t *)(void*)lParam1;
	int nbytes = (int)lParam2;

	if (pconfig == NULL) // set default state
	{	
		default_configuration(&hdrv->config);
		return ICERR_OK;
	}

	if (nbytes != sizeof(configvars_t))
		return ICERR_BADPARAM;

	hdrv->config = *pconfig;

	return ICERR_OK;
}


LONG my_icm_getinfo(void *hnd, LPARAM lParam1, LPARAM lParam2)
{
	driver_context_t * const hdrv = (driver_context_t*)hnd;
	if (hdrv == NULL)
		return ICERR_BADHANDLE;

	ICINFO* icinfo = (ICINFO*)lParam1;
	DWORD dwSize = (DWORD)lParam2;

	if (icinfo == NULL) 
		return sizeof(ICINFO);

	if (dwSize < sizeof(ICINFO)) 
		return 0;

	DWORD fourcc = dvfmt_to_fourcc(hdrv->dvfmt);

	icinfo->dwSize            = sizeof(ICINFO);
	icinfo->fccType           = ICTYPE_VIDEO;
	icinfo->fccHandler        = fourcc;
	icinfo->dwFlags           = 0;
	icinfo->dwVersion         = 0;
	icinfo->dwVersionICM      = ICVERSION;
	wcscpy_s(icinfo->szName, 16, CODEC_NAME);

	wcscpy_s(icinfo->szDescription, 128, 
		(hdrv->dvfmt == DVFORMAT_DV25) ? CODEC_DESCRIPTION_DV25 :
		(hdrv->dvfmt == DVFORMAT_DV50) ? CODEC_DESCRIPTION_DV50 :
		CODEC_DESCRIPTION_DVSD);

	return (DWORD)lParam2;
}


/****************************************
	compress messages
****************************************/

LONG my_icm_compress_query(void *hnd, LPARAM lParam1, LPARAM lParam2)
{
	driver_context_t * const hdrv = (driver_context_t*)hnd;
	if (hdrv == NULL)
		return ICERR_BADHANDLE;

	BITMAPINFOHEADER * const pbihInp = (BITMAPINFOHEADER*)lParam1;
	BITMAPINFOHEADER * const pbihOut = (BITMAPINFOHEADER*)lParam2;
	
	if (!pbihInp)
		return ICERR_BADFORMAT;

	if (get_input_colorformat(&hdrv->config, pbihInp) < 0)
		return ICERR_BADFORMAT;

	if (!pbihOut) 
		return ICERR_OK;
		
	DWORD fourcc = dvfmt_to_fourcc(hdrv->dvfmt);
	DWORD fcc = pbihOut->biCompression;

	if (!is_equal_fourcc(fcc, fourcc))
		return ICERR_BADFORMAT;
		
	return ICERR_OK;
}


LONG my_icm_compress_get_format(void *hnd, LPARAM lParam1, LPARAM lParam2)
{
	driver_context_t * const hdrv = (driver_context_t*)hnd;
	if (hdrv == NULL)
		return ICERR_BADHANDLE;

	BITMAPINFOHEADER * const pbihInp = (BITMAPINFOHEADER*)lParam1;
	BITMAPINFOHEADER * const pbihOut = (BITMAPINFOHEADER*)lParam2;

	if (!pbihInp)
		return ICERR_BADFORMAT;

	if (!pbihOut) 
		return sizeof(BITMAPINFOHEADER);

	const int nchannels = (hdrv->dvfmt == DVFORMAT_DV50) ? 2 : 1;

	if (ABS(pbihInp->biHeight) < 576)
	{
		pbihOut->biHeight = 480;
		pbihOut->biSizeImage = 120000 * nchannels;
	}
	else
	{
		pbihOut->biHeight = 576;
		pbihOut->biSizeImage = 144000 * nchannels;
	}

	pbihOut->biSize = sizeof(BITMAPINFOHEADER);
	pbihOut->biWidth = 720;
	pbihOut->biPlanes = 1;
	pbihOut->biBitCount = 24;
	pbihOut->biCompression = dvfmt_to_fourcc(hdrv->dvfmt);
    pbihOut->biXPelsPerMeter = 0; //117*100; 
    pbihOut->biYPelsPerMeter = 0; //128*100; 
    pbihOut->biClrUsed = 0; 
    pbihOut->biClrImportant = 0; 

	return ICERR_OK;
}


LONG my_icm_get_size(void *hnd, LPARAM lParam1)
{
	driver_context_t * const hdrv = (driver_context_t*)hnd;
	if (hdrv == NULL)
		return ICERR_BADHANDLE;

	BITMAPINFOHEADER* const pbihInp = (BITMAPINFOHEADER*)lParam1;

	// 1 DV-Frame = 12 (or 10 for NTSC) DIF-Sequence  150 DIF-Blocks  80 Bytes = 144000 Bytes (or 12000 for NTSC)
	// DVCPRO50 is twice big

	const int nchannels = (hdrv->dvfmt == DVFORMAT_DV50) ? 2 : 1;
	
	int nsequences = 12;

	if (pbihInp)
		nsequences = ABS(pbihInp->biHeight) >= 576 ? 12 : 10;

	return  150 * 80 * nsequences * nchannels;
}


LONG my_icm_compress_begin(void *hnd, LPARAM lParam1, LPARAM lParam2)
{
	driver_context_t * const hdrv = (driver_context_t*)hnd;
	if (hdrv == NULL)
		return ICERR_BADHANDLE;

	BITMAPINFOHEADER * const pbihInp = (BITMAPINFOHEADER*)lParam1;
	BITMAPINFOHEADER * const pbihOut = (BITMAPINFOHEADER*)lParam2;

	if ((hdrv == NULL) || (pbihInp == NULL) || (pbihOut == NULL))
		return ICERR_BADFORMAT;

	if (pbihInp->biWidth != 720)
		return ICERR_BADFORMAT;

	int height = ABS(pbihInp->biHeight);

	if (height != 480 &&
		height != 576)
	{
		return ICERR_BADFORMAT;
	}

	DWORD fcc = pbihOut->biCompression;
	DWORD fourcc = dvfmt_to_fourcc(hdrv->dvfmt);

	if (!is_equal_fourcc(fcc, fourcc))
		return ICERR_BADFORMAT;

	int is_PAL = (height == 576) ? 1 : 0;
	int color_format = get_input_colorformat(&hdrv->config, pbihInp);
	
	if (color_format <= 0)
		return ICERR_BADFORMAT;

	int flip_RGB = (pbihInp->biHeight < 0) ? 1 : 0;	// flip_RGB == 1 means first line in memory is top line of RGB-image

	if (hdrv->henc == NULL)
	{
		hdrv->henc = encoder_open();

		if (hdrv->henc == NULL)
			return ICERR_BADFORMAT;
	}

	encoder_configure(
		hdrv->henc, 
		color_format, 
		flip_RGB, 
		is_PAL,
		hdrv->dvfmt,
		&hdrv->config);

	return ICERR_OK;
}


LONG my_icm_compress(void *hnd, LPARAM lParam1, LPARAM lParam2)
{
	driver_context_t * const hdrv = (driver_context_t*)hnd;
	if (hdrv == NULL)
		return ICERR_BADHANDLE;

	ICCOMPRESS *icinfo = (ICCOMPRESS *)lParam1;
	DWORD dwSize = (DWORD)lParam2;
	
	if ((hdrv == NULL) || (icinfo == NULL))
		return ICERR_BADPARAM;

	if (hdrv->henc == NULL)
		return ICERR_BADPARAM;

	if (dwSize < sizeof(ICCOMPRESS))
		return ICERR_BADPARAM;

	if (icinfo->lpdwFlags)
		*icinfo->lpdwFlags = AVIIF_KEYFRAME;

	uint8_t const * inp = (uint8_t*)icinfo->lpInput;
	uint8_t * out = (uint8_t*)icinfo->lpOutput;

	if ((inp == NULL) || (out == NULL))
		return ICERR_BADFORMAT;

	if ((icinfo->lpbiInput == NULL) || (icinfo->lpbiOutput == NULL))
		return ICERR_BADPARAM;

	unsigned int out_size = icinfo->lpbiOutput->biSizeImage;

	int res = encode_DV(
		hdrv->henc, 
		inp, 
		icinfo->lpbiInput->biSizeImage, 
		out, 
		&out_size);

	icinfo->lpbiOutput->biSizeImage = out_size;

	return (res == 0) ? ICERR_OK : ICERR_BADPARAM; 
}


LONG my_icm_compress_end(void *hnd)
{
	driver_context_t * const hdrv = (driver_context_t*)hnd;
	if (hdrv == NULL)
		return ICERR_BADHANDLE;

	if (hdrv)
	{
		if (hdrv->henc)
		{
			encoder_close(hdrv->henc);
			hdrv->henc = NULL;
		}
	}

	return ICERR_OK;
}


/****************************************
	decompress messages
****************************************/

LONG my_icm_decompress_query(void *hnd, LPARAM lParam1, LPARAM lParam2)
{
	driver_context_t * const hdrv = (driver_context_t*)hnd;
	if (hdrv == NULL)
		return ICERR_BADHANDLE;

	BITMAPINFOHEADER * const pbihInp = (BITMAPINFOHEADER*)lParam1;
	BITMAPINFOHEADER * const pbihOut = (BITMAPINFOHEADER*)lParam2;
	
	if (!pbihInp) 
		return ICERR_BADFORMAT;

	DWORD fcc = pbihInp->biCompression;
	DWORD fourcc = dvfmt_to_fourcc(hdrv->dvfmt);

	if (!is_equal_fourcc(fcc, fourcc))
		return ICERR_BADFORMAT;

	if (!pbihOut) 
		return ICERR_OK;

	if (get_output_colorformat(&hdrv->config, pbihOut) < 0) 
		return ICERR_BADFORMAT;
		
	return ICERR_OK;
}


LONG my_icm_decompress_get_format(void *hnd, LPARAM lParam1, LPARAM lParam2)
{
	driver_context_t * const hdrv = (driver_context_t*)hnd;
	if (hdrv == NULL)
		return ICERR_BADHANDLE;

	BITMAPINFOHEADER * const pbihInp = (BITMAPINFOHEADER*)lParam1;
	BITMAPINFOHEADER * const pbihOut = (BITMAPINFOHEADER*)lParam2;

	if (pbihInp == NULL)
		return ICERR_BADFORMAT;

	if (pbihOut == NULL) 
		return sizeof(BITMAPINFOHEADER);

	DWORD fcc = pbihInp->biCompression;
	DWORD fourcc = dvfmt_to_fourcc(hdrv->dvfmt);

	if (!is_equal_fourcc(fcc, fourcc))
		return ICERR_BADFORMAT;

	const int width = 720;
	const int height = (ABS(pbihInp->biHeight) < 576) ? 480 : 576;
	const int npixel = width * height;

	memset(pbihOut, 0x00, sizeof(BITMAPINFOHEADER));

	pbihOut->biSize = sizeof(BITMAPINFOHEADER);
	pbihOut->biWidth = width;
	pbihOut->biPlanes = 1;
	pbihOut->biHeight = height;

	if (hdrv->config.decoder_output_format & COLORFORMAT_RGB32)
	{   // RGB32
		pbihOut->biBitCount = 32;    
		pbihOut->biCompression = BI_RGB;
		pbihOut->biSizeImage = npixel * 4;

		return ICERR_OK;
	}

	if (hdrv->config.decoder_output_format & COLORFORMAT_RGB24)
	{   // RGB24
		pbihOut->biBitCount = 24;    
		pbihOut->biCompression = BI_RGB;
		pbihOut->biSizeImage = npixel * 3;

		return ICERR_OK;
	}

	if (hdrv->config.decoder_output_format & COLORFORMAT_YV12)
	{	// YV12
		pbihOut->biBitCount = 12;       
		pbihOut->biCompression = FOURCC_YV12;
		pbihOut->biSizeImage = npixel * 3 / 2;

		return ICERR_OK;
	}

	if (hdrv->config.decoder_output_format & COLORFORMAT_YUY2)
	{	// YUY2
		pbihOut->biBitCount = 16;       // YUY2
		pbihOut->biCompression = FOURCC_YUY2;
		pbihOut->biSizeImage = npixel * 2;

		return ICERR_OK;
	}

	// RGB24
	pbihOut->biBitCount = 24;    
	pbihOut->biCompression = BI_RGB;
	pbihOut->biSizeImage = npixel * 3;

	return ICERR_OK;
}


LONG my_icm_decompress_begin(void *hnd, LPARAM lParam1, LPARAM lParam2)
{
	driver_context_t * const hdrv = (driver_context_t*)hnd;
	if (hdrv == NULL)
		return ICERR_BADHANDLE;

	BITMAPINFOHEADER * const pbihInp = (BITMAPINFOHEADER*)lParam1;
	BITMAPINFOHEADER * const pbihOut = (BITMAPINFOHEADER*)lParam2;

	if ((hdrv == NULL) || (pbihInp == NULL) || (pbihOut == NULL))
		return ICERR_BADFORMAT;

	DWORD fcc = pbihInp->biCompression;
	DWORD fourcc = dvfmt_to_fourcc(hdrv->dvfmt);

	if (!is_equal_fourcc(fcc, fourcc))
		return ICERR_BADFORMAT;
		
	if (pbihOut->biWidth != 720)
		return ICERR_BADFORMAT;

	const int height = ABS(pbihOut->biHeight);

	if (height != 480 &&
		height != 576)
	{
		return ICERR_BADFORMAT;
	}

	int is_PAL = (height == 576) ? 1 : 0;
	int color_format = get_output_colorformat(&hdrv->config, pbihOut);

	if (color_format <= 0)
		return ICERR_BADFORMAT;

	int flip_RGB = (pbihOut->biHeight < 0);	// flip_RGB == 1 means first line in memory is top line of RGB-image

	if (hdrv->hdec == NULL)
	{
		hdrv->hdec = decoder_open();

		if (hdrv->hdec == NULL)
			return ICERR_BADFORMAT;
	}

	decoder_configure(
		hdrv->hdec, 
		color_format, 
		flip_RGB, 
		is_PAL, 
		hdrv->dvfmt,
		&hdrv->config);

	return ICERR_OK;
}


LONG my_icm_decompress(void *hnd, LPARAM lParam1, LPARAM lParam2)
{
	driver_context_t * const hdrv = (driver_context_t*)hnd;
	if (hdrv == NULL)
		return ICERR_BADHANDLE;

	ICDECOMPRESS * icinfo = (ICDECOMPRESS *)lParam1; 
	DWORD dwSize = (DWORD)lParam2;

	if ((hdrv == NULL) || (icinfo == NULL))
		return ICERR_BADPARAM;
	
	if (dwSize < sizeof(ICDECOMPRESS))
		return ICERR_BADPARAM;

	if (hdrv->hdec == NULL)
		return ICERR_BADPARAM;

	if (icinfo->dwFlags & ICDECOMPRESS_NULLFRAME)
		return ICERR_OK;
	
	uint8_t const * inp = (uint8_t*)icinfo->lpInput;
	uint8_t * out = (uint8_t*)icinfo->lpOutput;

	if ((inp == NULL) || (out == NULL))
		return ICERR_BADPARAM;

	if ((icinfo->lpbiInput == NULL) || (icinfo->lpbiOutput == NULL))
		return ICERR_BADPARAM;

	unsigned int out_size = icinfo->lpbiOutput->biSizeImage;

	int res = decode_DV(
		hdrv->hdec, 
		inp,
		icinfo->lpbiInput->biSizeImage,
		out,
		&out_size);

	return (res == 0) ? ICERR_OK : ICERR_BADPARAM;
}


LONG my_icm_decompress_end(void *hnd)
{
	driver_context_t * const hdrv = (driver_context_t*)hnd;
	if (hdrv == NULL)
		return ICERR_BADHANDLE;
	
	if (hdrv)
	{
		if (hdrv->hdec)
		{
			decoder_close(hdrv->hdec);
			hdrv->hdec = NULL;
		}
	}

	return ICERR_OK;
}


static const configvars_t myconfig_default = {
	QUALITY_MODE_NORMAL,	// encoder_quality_mode
	COLORFORMAT_YUY2 | COLORFORMAT_YV12 | COLORFORMAT_RGB32 | COLORFORMAT_RGB24,	// encoder_input_format;
	0,	// encoder_RGB_to_YUV_conversion = 0;
	YV12_CHROMASAMPLING_MPEG2_I,	// encoder_yv12_chroma_sampling;
	25,	// encoder_dct248_threshold = 10;
	15,	// encoder_dct248_counts = 15;
	0,	// aspect_ratio
	0,	// encoder_coring_enable = 0;
	{	// encoder_coring_matrix[64]
	    0,    0,    0,    0,    0,    0,    0,    0,
	    0,    0,    0,    0,    0,    0,    0,    0,
	    0,    0,    0,    0,    0,    0,    0, 1023,
	    0,    0,    0,    0,    0,    0, 1023, 1023,
	    0,    0,    0,    0,    0, 1023, 1023, 1023,
	    0,    0,    0,    0, 1023, 1023, 1023, 1023,
	    0,    0,    0, 1023, 1023, 1023, 1023, 1023,
	    0,    0, 1023, 1023, 1023, 1023, 1023, 1023
	},
	{  0, 10, 210 },	// encoder_classification_Y[3]
	{  0,  0, 3 },	// encoder_classification_V[3]
	{  0,  0, 3 },	// encoder_classification_U[3]
	COLORFORMAT_YUY2 | COLORFORMAT_YV12 | COLORFORMAT_RGB32 | COLORFORMAT_RGB24,	// decoder_output_format
	0,	// decoder_YUV_to_RGB_conversion = 0;
	YV12_CHROMASAMPLING_MPEG2_I	// decoder_YV12_chroma_sampling;
};


static const wchar_t * const quality_mode_names[NO_OF_QUALITY_MODES] = {
	L"normal", L"fast" };


static int default_configuration(configvars_t *pconfig)
{
	memcpy(pconfig, &myconfig_default, sizeof(configvars_t));
	return 0;
}


static int load_configuration(configvars_t *pconfig, dvformat_t dvfmt)
{
	memcpy(pconfig, &myconfig_default, sizeof(configvars_t));

	HKEY hKey = NULL;

	LPCWSTR const psubkey =
		(dvfmt == DVFORMAT_DV25) ? SETTINGS_REG_NAME_DV25 :
		(dvfmt == DVFORMAT_DV50) ? SETTINGS_REG_NAME_DV50 :
		SETTINGS_REG_NAME_DVSD;

	LONG res = RegOpenKeyExW(
		SETTINGS_REG_KEY,
		psubkey, 
		0, 
		KEY_READ, 
		&hKey);

	if (res != ERROR_SUCCESS)
		return -1;

	DWORD nbytes_config = sizeof(*pconfig);
	DWORD dwtype = 0;

	res = RegQueryValueExW(
		hKey,
		SETTINGS_REG_VAL,
		0,
		&dwtype,
		(BYTE*)pconfig,
		&nbytes_config);

	RegCloseKey(hKey);

	if (pconfig->encoder_quality_mode >= NO_OF_QUALITY_MODES)
		pconfig->encoder_quality_mode = QUALITY_MODE_NORMAL;

	if (res != ERROR_SUCCESS ||
		dwtype != REG_BINARY ||
		nbytes_config != sizeof(*pconfig))
	{
		return -1;
	}

	return 0;
}


static int save_configuration(configvars_t const *pconfig, dvformat_t dvfmt)
{
	HKEY hKey = NULL;
	DWORD dwDisposition = 0;

	LPCWSTR const psubkey =
		(dvfmt == DVFORMAT_DV25) ? SETTINGS_REG_NAME_DV25 :
		(dvfmt == DVFORMAT_DV50) ? SETTINGS_REG_NAME_DV50 :
		SETTINGS_REG_NAME_DVSD;

	LONG res = RegCreateKeyExW(
		SETTINGS_REG_KEY,
		psubkey,
		0,
		NULL,
		REG_OPTION_NON_VOLATILE,
		KEY_WRITE,
		0,
		&hKey,
		&dwDisposition);
	
	if (res != ERROR_SUCCESS)
		return -1;

	res = RegSetValueExW(
		hKey,
		SETTINGS_REG_VAL,
		0,
		REG_BINARY,
		(BYTE*)pconfig,
		sizeof(*pconfig));

	RegCloseKey(hKey);

	return (res == ERROR_SUCCESS) ? 0 : -1;
}


static void redraw_dialog(int reset, HWND hwndDlg, const configvars_t *pconfig)
{
	if (reset)
	{
		HWND hctl_quality_mode = GetDlgItem(hwndDlg, IDC_QUALITY_MODE);

		SendMessage(hctl_quality_mode, CB_SETCURSEL, pconfig->encoder_quality_mode, 0);

		// input format
		CheckDlgButton(hwndDlg, IDC_IF_RGB24, (pconfig->encoder_input_format&COLORFORMAT_RGB24) ? BST_CHECKED : BST_UNCHECKED);
		CheckDlgButton(hwndDlg, IDC_IF_RGB32, (pconfig->encoder_input_format&COLORFORMAT_RGB32) ? BST_CHECKED : BST_UNCHECKED);
		CheckDlgButton(hwndDlg, IDC_IF_YUY2,  (pconfig->encoder_input_format&COLORFORMAT_YUY2)  ? BST_CHECKED : BST_UNCHECKED);
		CheckDlgButton(hwndDlg, IDC_IF_YV12,  (pconfig->encoder_input_format&COLORFORMAT_YV12)  ? BST_CHECKED : BST_UNCHECKED);
		
		// rgb-to-yuv conversion
		CheckDlgButton(hwndDlg, IDC_RGB2YUV_601,       (pconfig->encoder_RGB_to_YUV_conversion == RGB2YUV_601)       ? BST_CHECKED : BST_UNCHECKED);
		CheckDlgButton(hwndDlg, IDC_RGB2YUV_UNCHANGED, (pconfig->encoder_RGB_to_YUV_conversion == RGB2YUV_UNCHANGED) ? BST_CHECKED : BST_UNCHECKED);

		// input yv12 chroma sampling
		CheckDlgButton(hwndDlg, IDC_YV12_CHROMA_IN_DV,                (pconfig->encoder_YV12_chroma_sampling == YV12_CHROMASAMPLING_DV)      ? BST_CHECKED : BST_UNCHECKED);
		CheckDlgButton(hwndDlg, IDC_YV12_CHROMA_IN_MPEG2_PROGRESSIVE, (pconfig->encoder_YV12_chroma_sampling == YV12_CHROMASAMPLING_MPEG2_P) ? BST_CHECKED : BST_UNCHECKED);
		CheckDlgButton(hwndDlg, IDC_YV12_CHROMA_IN_MPEG2_INTERLACED,  (pconfig->encoder_YV12_chroma_sampling == YV12_CHROMASAMPLING_MPEG2_I) ? BST_CHECKED : BST_UNCHECKED);

		// aspect ratio
		CheckDlgButton(hwndDlg, IDC_ASPECTRATIO_4TO3,  (pconfig->encoder_aspect_ratio==ASPECTRATIO_4TO3) ? BST_CHECKED : BST_UNCHECKED);
		CheckDlgButton(hwndDlg, IDC_ASPECTRATIO_16TO9, (pconfig->encoder_aspect_ratio!=ASPECTRATIO_4TO3) ? BST_CHECKED : BST_UNCHECKED);

		// output format
		CheckDlgButton(hwndDlg, IDC_OF_RGB24, (pconfig->decoder_output_format&COLORFORMAT_RGB24) ? BST_CHECKED : BST_UNCHECKED);
		CheckDlgButton(hwndDlg, IDC_OF_RGB32, (pconfig->decoder_output_format&COLORFORMAT_RGB32) ? BST_CHECKED : BST_UNCHECKED);
		CheckDlgButton(hwndDlg, IDC_OF_YUY2,  (pconfig->decoder_output_format&COLORFORMAT_YUY2)  ? BST_CHECKED : BST_UNCHECKED);
		CheckDlgButton(hwndDlg, IDC_OF_YV12,  (pconfig->decoder_output_format&COLORFORMAT_YV12)  ? BST_CHECKED : BST_UNCHECKED);

		// yuv-to-rgb conversion
		CheckDlgButton(hwndDlg, IDC_YUY2RGB_601,       (pconfig->decoder_YUV_to_RGB_conversion == RGB2YUV_601)       ? BST_CHECKED : BST_UNCHECKED);
		CheckDlgButton(hwndDlg, IDC_YUY2RGB_UNCHANGED, (pconfig->decoder_YUV_to_RGB_conversion == RGB2YUV_UNCHANGED) ? BST_CHECKED : BST_UNCHECKED);

		// output yv12 chroma sampling
		CheckDlgButton(hwndDlg, IDC_YV12_CHROMA_OUT_DV,                (pconfig->decoder_YV12_chroma_sampling == YV12_CHROMASAMPLING_DV)      ? BST_CHECKED : BST_UNCHECKED);
		CheckDlgButton(hwndDlg, IDC_YV12_CHROMA_OUT_MPEG2_PROGRESSIVE, (pconfig->decoder_YV12_chroma_sampling == YV12_CHROMASAMPLING_MPEG2_P) ? BST_CHECKED : BST_UNCHECKED);
		CheckDlgButton(hwndDlg, IDC_YV12_CHROMA_OUT_MPEG2_INTERLACED,  (pconfig->decoder_YV12_chroma_sampling == YV12_CHROMASAMPLING_MPEG2_I) ? BST_CHECKED : BST_UNCHECKED);
	}

	int is_rgb_input = 
		(IsDlgButtonChecked(hwndDlg, IDC_IF_RGB24) == BST_CHECKED) ||
		(IsDlgButtonChecked(hwndDlg, IDC_IF_RGB32) == BST_CHECKED);

	int is_yv12_input = (IsDlgButtonChecked(hwndDlg, IDC_IF_YV12) == BST_CHECKED);

	EnableWindow(GetDlgItem(hwndDlg, IDC_RGB2YUV_601), is_rgb_input);
	EnableWindow(GetDlgItem(hwndDlg, IDC_RGB2YUV_UNCHANGED), is_rgb_input);

	EnableWindow(GetDlgItem(hwndDlg, IDC_YV12_CHROMA_IN_DV), is_yv12_input);
	EnableWindow(GetDlgItem(hwndDlg, IDC_YV12_CHROMA_IN_MPEG2_PROGRESSIVE), is_yv12_input);
	EnableWindow(GetDlgItem(hwndDlg, IDC_YV12_CHROMA_IN_MPEG2_INTERLACED), is_yv12_input);

	int is_rgb_output = 
		(IsDlgButtonChecked(hwndDlg, IDC_OF_RGB24) == BST_CHECKED) ||
		(IsDlgButtonChecked(hwndDlg, IDC_OF_RGB32) == BST_CHECKED);

	int is_yv12_output = (IsDlgButtonChecked(hwndDlg, IDC_OF_YV12) == BST_CHECKED);

	EnableWindow(GetDlgItem(hwndDlg, IDC_YUY2RGB_601), is_rgb_output);
	EnableWindow(GetDlgItem(hwndDlg, IDC_YUY2RGB_UNCHANGED), is_rgb_output);

	EnableWindow(GetDlgItem(hwndDlg, IDC_YV12_CHROMA_OUT_DV), is_yv12_output);
	EnableWindow(GetDlgItem(hwndDlg, IDC_YV12_CHROMA_OUT_MPEG2_PROGRESSIVE), is_yv12_output);
	EnableWindow(GetDlgItem(hwndDlg, IDC_YV12_CHROMA_OUT_MPEG2_INTERLACED), is_yv12_output);
}


#define LIMIT_MIN_MAX(x, xmin, xmax) (((x)<xmin) ? xmin : (((x)>xmax) ? xmax : (x)))

static INT_PTR CALLBACK ConfigureDialogProc(HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam) 
{
	configvars_t * pconfig = (configvars_t*)GetWindowLongPtr(hwndDlg, DWLP_USER);

	if (pconfig == NULL) 
	{
		if (uMsg != WM_INITDIALOG)
			return FALSE;

		pconfig = (configvars_t*)lParam;
		SetWindowLongPtr(hwndDlg, DWLP_USER, (LONG_PTR)pconfig);
	}

	if (uMsg == WM_INITDIALOG) 
	{
		// quality mode
		HWND hctl_quality_mode = GetDlgItem(hwndDlg, IDC_QUALITY_MODE);

		if (SendMessage(hctl_quality_mode, CB_GETCOUNT, 0, 0) == 0)
		{
			for (int i = 0; i < NO_OF_QUALITY_MODES; i++)
			  SendMessageW(hctl_quality_mode, CB_ADDSTRING, 0, (LPARAM)quality_mode_names[i]);
		}

		redraw_dialog(TRUE, hwndDlg, pconfig);
    }
	else if (uMsg == WM_COMMAND) 
	{
		redraw_dialog(FALSE, hwndDlg, pconfig);

		switch (LOWORD(wParam)) {

		case IDOK:

			// quality-mode
			pconfig->encoder_quality_mode = (int)SendMessage(GetDlgItem(hwndDlg, IDC_QUALITY_MODE), CB_GETCURSEL, 0, 0);

			// input format
			pconfig->encoder_input_format = 0;

			if (IsDlgButtonChecked(hwndDlg, IDC_IF_RGB24) == BST_CHECKED) pconfig->encoder_input_format |= COLORFORMAT_RGB24;
			if (IsDlgButtonChecked(hwndDlg, IDC_IF_RGB32) == BST_CHECKED) pconfig->encoder_input_format |= COLORFORMAT_RGB32;
			if (IsDlgButtonChecked(hwndDlg, IDC_IF_YUY2)  == BST_CHECKED) pconfig->encoder_input_format |= COLORFORMAT_YUY2;
			if (IsDlgButtonChecked(hwndDlg, IDC_IF_YV12)  == BST_CHECKED) pconfig->encoder_input_format |= COLORFORMAT_YV12;

			// rgb-to-yuv conversion
			if (IsDlgButtonChecked(hwndDlg, IDC_RGB2YUV_601) == BST_CHECKED)
				pconfig->encoder_RGB_to_YUV_conversion = RGB2YUV_601;
			else 
				pconfig->encoder_RGB_to_YUV_conversion = RGB2YUV_UNCHANGED;

			// input yv12 chroma sampling
			if (IsDlgButtonChecked(hwndDlg, IDC_YV12_CHROMA_IN_DV) == BST_CHECKED)
				pconfig->encoder_YV12_chroma_sampling = YV12_CHROMASAMPLING_DV;
			else if (IsDlgButtonChecked(hwndDlg, IDC_YV12_CHROMA_IN_MPEG2_PROGRESSIVE) == BST_CHECKED)
				pconfig->encoder_YV12_chroma_sampling = YV12_CHROMASAMPLING_MPEG2_P;
			else 
				pconfig->encoder_YV12_chroma_sampling = YV12_CHROMASAMPLING_MPEG2_I;

			// aspect ratio
			if (IsDlgButtonChecked(hwndDlg, IDC_ASPECTRATIO_4TO3) == BST_CHECKED)
				pconfig->encoder_aspect_ratio = ASPECTRATIO_4TO3;
			else 
				pconfig->encoder_aspect_ratio = ASPECTRATIO_16TO9;

			// output format
			pconfig->decoder_output_format = 0;

			if (IsDlgButtonChecked(hwndDlg, IDC_OF_RGB24) == BST_CHECKED) pconfig->decoder_output_format |= COLORFORMAT_RGB24;
			if (IsDlgButtonChecked(hwndDlg, IDC_OF_RGB32) == BST_CHECKED) pconfig->decoder_output_format |= COLORFORMAT_RGB32;
			if (IsDlgButtonChecked(hwndDlg, IDC_OF_YUY2)  == BST_CHECKED) pconfig->decoder_output_format |= COLORFORMAT_YUY2;
			if (IsDlgButtonChecked(hwndDlg, IDC_OF_YV12)  == BST_CHECKED) pconfig->decoder_output_format |= COLORFORMAT_YV12;

			// yuv-to-rgb conversion
			if (IsDlgButtonChecked(hwndDlg, IDC_YUY2RGB_601) == BST_CHECKED)
				pconfig->decoder_YUV_to_RGB_conversion = RGB2YUV_601;
			else 
				pconfig->decoder_YUV_to_RGB_conversion = RGB2YUV_UNCHANGED;

			// output yv12 chroma sampling
			if (IsDlgButtonChecked(hwndDlg, IDC_YV12_CHROMA_OUT_DV) == BST_CHECKED)
				pconfig->decoder_YV12_chroma_sampling = YV12_CHROMASAMPLING_DV;
			else if (IsDlgButtonChecked(hwndDlg, IDC_YV12_CHROMA_OUT_MPEG2_PROGRESSIVE) == BST_CHECKED)
				pconfig->decoder_YV12_chroma_sampling = YV12_CHROMASAMPLING_MPEG2_P;
			else 
				pconfig->decoder_YV12_chroma_sampling = YV12_CHROMASAMPLING_MPEG2_I;

			EndDialog(hwndDlg, 1);
			break;

		case IDCANCEL:
			EndDialog(hwndDlg, 0);
			break;

		case IDDEFAULT:
			redraw_dialog(TRUE, hwndDlg, &myconfig_default);
			break;

		default:
			break;
		}
	}

	return FALSE;
}
