/******************************************************************************/
/*                                                                            */
/*  Cedocida DV-Codec                                                         */
/*                                                                            */
/*  Copyright (C) 2005 Andreas Dittrich <dittrich@cithraidt.de>               */
/*  Project start: 15.12.2003                                                 */
/*  Last change  : 15.06.2007                                                 */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/*  This program is free software; you can redistribute it and/or modify      */
/*  it under the terms of the GNU General Public License as published by      */
/*  the Free Software Foundation; either version 2 of the License, or         */
/*  (at your option) any later version.                                       */
/*                                                                            */
/*  This program is distributed in the hope that it will be useful,           */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             */
/*  GNU General Public License for more details.                              */
/*                                                                            */
/*  You should have received a copy of the GNU General Public License         */
/*  along with this program; if not, write to the Free Software               */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
/*                                                                            */
/******************************************************************************/

#include <stdlib.h>
#include <string.h>

#include "inttypes.h"

#include "signalproc.h"
#include "dv_spec.h"
#include "dct/dct.h"
#include "vlc/decode_vlc.h"
#include "decode_tables.h"
#include "config.h"
#include "decode.h"

#include "convert/convert.h"


#define LIMIT_1_TO_254(x) (((x)<1) ? 1 : (((x)>254) ? 254 : (x)))
#define LIMIT_2_TO_508(x) (((x)<2) ? 2 : (((x)>508) ? 508 : (x)))
#define LIMIT_MINUS255_TO_PLUS255(x) (((x)<-255) ? (-255) : (((x)>255) ? (255) : (x)))


typedef struct {
	uint8_t *buffer_base;
	int16_t *buffer_5_MBs;
	uint8_t *buffer_frame;

	int color_format;
	int flip_RGB;
	int is_system_625;
	dvformat_t dvfmt;

	int nbytes_dv;
	int nbytes_frame;

	configvars_t config;
} decoder_context_t;


static void decode_420_DV(
	void *hnd, 
	uint8_t const *src_DV, 
	uint8_t *dst);

static void decode_422_DV(
	void *hnd, 
	uint8_t const *src_DV, 
	uint8_t *dst);

static void decode_411_DV(
	void *hnd, 
	uint8_t const *src_DV, 
	uint8_t *dst);

static void decode_DIF_videosegment(
	uint8_t const *DIF_block, 
	int16_t *dst_5_MBs, 
	int is_MB211);

static void deinterleave_MB_YV12_PAL(
	int16_t const *src_MB, 
	uint8_t *dst_YUV420_Y, 
	uint8_t *dst_YUV420_U, 
	uint8_t *dst_YUV420_V, 
	int dst_YUV420_Y_pitch, 
	int dst_YUV420_UV_pitch);

static void deinterleave_MB211_YUY2(
	int16_t const *src_MB, 
	uint8_t *dst_YUY2, 
	int dst_YUY2_stride);

static void deinterleave_MB_YV12_NTSC(
	int16_t const *src_MB, 
	uint8_t *dst_YUV420_Y, 
	uint8_t *dst_YUV420_U, 
	uint8_t *dst_YUV420_V, 
	int dst_YUV420_Y_pitch, 
	int dst_YUV420_UV_pitch);

static void deinterleave_MBborder_YV12_NTSC(
	int16_t const *src_MB, 
	uint8_t *dst_YUV420_Y, 
	uint8_t *dst_YUV420_U, 
	uint8_t *dst_YUV420_V, 
	int dst_YUV420_Y_pitch, 
	int dst_YUV420_UV_pitch);

static void YUV420_to_YUY2(
	uint8_t const *src_YUV420_Y, 
	uint8_t const *src_YUV420_U, 
	uint8_t const *src_YUV420_V, 
	uint8_t *dst_YUY2, 
	int src_YUV420_Y_pitch, 
	int src_YUV420_UV_pitch, 
	int dst_YUY2_pitch, 
	int width, 
	int height);

static void YUV411_to_YUY2(
	uint8_t const *src_YUV420_Y, 
	uint8_t const *src_YUV420_U, 
	uint8_t const *src_YUV420_V, 
	uint8_t *dst_YUY2, 
	int src_YUV420_Y_pitch, 
	int src_YUV420_UV_pitch, 
	int dst_YUY2_pitch, 
	int width, 
	int height);


#define LOG2_ALIGN 4
#define SIZE_BUFFER_5_MBS (5*6*64*2)
#define SIZE_BUFFER_FRAME (720*576*4)

void * decoder_open()
{
	decoder_context_t * const h = (decoder_context_t*)malloc(sizeof(decoder_context_t));
	if (h == NULL)
		return NULL;

	memset(h, 0x00, sizeof(decoder_context_t));

	h->buffer_base = (uint8_t*)malloc(((1 << LOG2_ALIGN) - 1) + SIZE_BUFFER_FRAME + SIZE_BUFFER_5_MBS);
	if (h->buffer_base == NULL)
		decoder_close(h);

	h->buffer_frame = h->buffer_base + ((int)((uint8_t*)NULL - h->buffer_base) & ((1 << LOG2_ALIGN) - 1));
	h->buffer_5_MBs = (int16_t*)(h->buffer_frame + SIZE_BUFFER_FRAME);

	h->color_format = COLORFORMAT_YV12;

	return h;
}

void decoder_configure(
	void *hnd, 
	int color_format, 
	int flip_RGB, 
	int is_system_625, 
	dvformat_t dvfmt,
	configvars_t const *pconfig)
{
	decoder_context_t * const h = (decoder_context_t*)hnd;
	if (h == NULL)
		return;

	h->color_format = color_format;
	h->flip_RGB = flip_RGB;
	h->is_system_625 = is_system_625;
	h->dvfmt = dvfmt;

	int const nsequences = is_system_625 ? 12 : 10;
	int const nchannel = (dvfmt == DVFORMAT_DV50) ? 2 : 1;
	int const width = 720;
	int const height = is_system_625 ? 576 : 480;

	h->nbytes_dv = nchannel * nsequences * 150 * 80;

	h->nbytes_frame = height *
		(h->color_format == COLORFORMAT_YV12)  ? width * height * 3 / 2 :
		(h->color_format == COLORFORMAT_YUY2)  ? width * height * 2 :
		(h->color_format == COLORFORMAT_RGB24) ? (width * height * 3 + 3) & ~0x03  :
		(h->color_format == COLORFORMAT_RGB32) ? width * height * 4 :
		width * height * 4;

	memcpy(&h->config, pconfig, sizeof(configvars_t));
}


void decoder_close(void *hnd)
{
	decoder_context_t * const h = (decoder_context_t*)hnd;
	if (h == NULL)
		return;

	if (h->buffer_base!=NULL)
		free(h->buffer_base);

	free(h);
}


int decode_DV(
	void *hnd, 
	uint8_t const *src_DV, 
	unsigned int size_src, 
	uint8_t *dst, 
	unsigned int *p_size_dst)
{
	decoder_context_t * const h = (decoder_context_t*)hnd;
	
	if (h == NULL || p_size_dst == NULL)
		return -1;

	if ((int)size_src < h->nbytes_dv ||
		(int)*p_size_dst < h->nbytes_frame)
	{
		return -2;
	}

	*p_size_dst = h->nbytes_frame;
	
	switch (h->dvfmt) 
	{
	case DVFORMAT_DVSD:
	{
		if (h->is_system_625)
		{
			decode_420_DV(h, src_DV, dst);
		}
		else
		{
			decode_411_DV(h, src_DV, dst);
		}

		break;
	}

	case DVFORMAT_DV25:
		decode_411_DV(h, src_DV, dst);
		break;

	case DVFORMAT_DV50:
		decode_422_DV(h, src_DV, dst);
		break;

	default:
		; // do nothing
	}

	emms_asm();
	return 0;
}


static void decode_420_DV(void *hnd, uint8_t const *src_DV, uint8_t *dst)
{
	decoder_context_t * const h = (decoder_context_t*)hnd;
	
	const int width = 720;
	const int height = 576;
	const int nsequences = 12;
	int color_format = h->color_format;

	int i, r, j, n;
	int16_t * dst_5_MBs = h->buffer_5_MBs;
	uint8_t * dst_step1 = dst;
	uint8_t * dst_step2 = h->buffer_frame;

	if (color_format == COLORFORMAT_YUY2)
	{
		dst_step1 = h->buffer_frame;
	}

	for (i = 0; i < nsequences; i++)	// sequence
	{
		for (r = 0; r < 3; r++)	// row_MB in SB
		{
			for (j = 0; j < 9; j++)	// col_MB in SB
			{
				int block_indx = 5 * SB_row_order[r][j]; 
				int DBN_1st = 7 + block_indx + block_indx / 15;	// first DIF_blocknumber
				uint8_t const * src_DIF_videosegment = src_DV + i*(150*80) + DBN_1st*(80);

				decode_DIF_videosegment(src_DIF_videosegment, dst_5_MBs, 0);

				for (n = 0; n < 5; n++)		// col_SB
				{
					// save it first as YV12 and then convert to other formats
					// todo: direct conversion to the right format
					int16_t * src_MB = dst_5_MBs + n*6*64;

					int dst_Y_pitch = width;
					int dst_UV_pitch = dst_Y_pitch >> 1;

					uint8_t * dst_Y = dst_step1; 
					uint8_t * dst_V = dst_Y + height * dst_Y_pitch; 
					uint8_t * dst_U = dst_V + (height >> 1) * dst_UV_pitch;

					static const int modulo_offset[5] = { 2, 6, 8, 0, 4 };
					static const int j_SB[5] = { 2, 1, 3, 0, 4 };

					int row_SB = (i + modulo_offset[n]) % nsequences;

					uint8_t * dst_YUV420_Y = dst_Y + (j_SB[n])*(9*16) + j*(16) + (row_SB*3+r)*16*dst_Y_pitch;
					uint8_t * dst_YUV420_V = dst_V + (j_SB[n])*(9*8)  + j*(8)  + (row_SB*3+r)*8*dst_UV_pitch;
					uint8_t * dst_YUV420_U = dst_U + (j_SB[n])*(9*8)  + j*(8)  + (row_SB*3+r)*8*dst_UV_pitch;

					deinterleave_MB_YV12_PAL(
						src_MB, 
						dst_YUV420_Y, 
						dst_YUV420_U, 
						dst_YUV420_V, 
						dst_Y_pitch, 
						dst_UV_pitch);
				}
			}
		}
	}

	if (color_format == COLORFORMAT_YV12)
	{
		if (h->config.decoder_YV12_chroma_sampling != YV12_CHROMASAMPLING_DV)
		{
			YUV420_to_YUY2(
				dst_step1, 
				dst_step1 + height*width + (height >> 1)*(width >> 1), 
				dst_step1 + height*width, 
				dst_step2, 
				width, 
				width >> 1,
				width << 1, 
				width, 
				height);

			if (h->config.decoder_YV12_chroma_sampling == YV12_CHROMASAMPLING_MPEG2_I)
			{
				mmx_yuy2_i_to_yv12(
					dst_step2,
					width*2,
					width*2,
					dst,
					dst + height*width + (height >> 1)*(width >> 1), 
					dst + height*width, 
					width,
					width >> 1,
					height);
			}
			else
			{
				mmx_yuy2_to_yv12((
					uint8_t*)dst_step2,
					width*2,
					width*2, 
					(uint8_t*)dst, 
					(uint8_t*)dst + height*width + (height >> 1)*(width >> 1),
					(uint8_t*)dst + height*width,
					width,
					width >> 1,
					height);
			}
		}
	}
	else if (color_format == COLORFORMAT_YUY2)
	{
		YUV420_to_YUY2(
			dst_step1, 
			dst_step1 + height*width + (height >> 1)*(width >> 1), 
			dst_step1 + height*width, 
			dst, 
			width, 
			width >> 1,
			width << 1, 
			width, 
			height);
	}
	else if (color_format == COLORFORMAT_RGB32)
	{
		YUV420_to_YUY2(
			dst_step1, 
			dst_step1 + height*width + (height >> 1)*(width >> 1), 
			dst_step1 + height*width,
			dst_step2, 
			width, 
			width >> 1,
			width << 1, 
			width, 
			height);

		const int is_RGB2YUV_FULL = (h->config.decoder_YUV_to_RGB_conversion == RGB2YUV_601 ? 0 : 1);

		if (h->flip_RGB)
		{
			mmx_YUY2toRGB32(
				dst_step2 + (height-1)*width*2, 
				dst, 
				dst_step2 - width*2, 
				-width*2, 
				width*2, 
				is_RGB2YUV_FULL);
		}
		else
		{
			mmx_YUY2toRGB32(
				dst_step2, 
				dst, 
				dst_step2 + width*height*2, 
				width*2, 
				width*2, 
				is_RGB2YUV_FULL);
		}
	}
	else if (color_format == COLORFORMAT_RGB24)
	{
		YUV420_to_YUY2(
			dst_step1, 
			dst_step1 + height*width + (height >> 1)*(width >> 1),
			dst_step1 + height*width, 
			dst_step2, 
			width, 
			width >> 1, 
			width << 1, 
			width, 
			height);

		const int is_RGB2YUV_FULL = (h->config.decoder_YUV_to_RGB_conversion == RGB2YUV_601 ? 0 : 1);

		if (h->flip_RGB)
		{
			mmx_YUY2toRGB24(
				dst_step2 + (height-1)*width*2, 
				dst, 
				dst_step2 - width*2, 
				-width*2, 
				width*2, 
				is_RGB2YUV_FULL);
		}
		else
		{
			mmx_YUY2toRGB24(
				dst_step2, 
				dst, 
				dst_step2 + width*height*2, 
				width*2, 
				width*2, 
				is_RGB2YUV_FULL);
		}
	}
}


static void decode_422_DV(void *hnd, uint8_t const *src_DV, uint8_t *dst)
{
	decoder_context_t * const h = (decoder_context_t*)hnd;
	
	const int width = 720;
	const int height = h->is_system_625 ? 576 : 480;
	const int nsequences = h->is_system_625 ? 12 : 10;
	int color_format = h->color_format;

	int i, r, j, n, ch;
	int16_t * dst_5_MBs = h->buffer_5_MBs;
	uint8_t * dst_step1 = (color_format == COLORFORMAT_YUY2) ? dst : h->buffer_frame;
	uint8_t * dst_step2 = h->buffer_frame;

	for (ch = 0; ch < 2; ch++)
	for (i = 0; i < nsequences; i++)	// sequence
	{
		for (r = 0; r < 3; r++)	// row_MB in SB
		{
			for (j = 0; j < 9; j++)	// col_MB in SB
			{
				int block_indx = 5 * SB_row_order[r][j]; 
				int DBN_1st = 7 + block_indx + block_indx / 15;	// first DIF_blocknumber
				uint8_t const * src_DIF_videosegment = src_DV + (ch*nsequences + i)*(150*80) + DBN_1st*(80);

				decode_DIF_videosegment(src_DIF_videosegment, dst_5_MBs, 1);

				for (n = 0; n < 5; n++)		// col_SB
				{
					// save it first as YV12 and then convert to other formats
					// todo: direct conversion to the right format
					int16_t * src_MB = dst_5_MBs + n*6*64;

					static const int modulo_offset[5] = { 2, 6, 8, 0, 4 };
					static const int j_SB[5] = { 2, 1, 3, 0, 4 };

					int row_SB = ((i + modulo_offset[n]) % nsequences) * 2 + ch;

					int dst_YUY2_stride = width << 1;
					uint8_t * dst_YUY2 = dst_step1 + (j_SB[n])*(9*16*2) + j*(16*2) + (row_SB*3+r)*8*dst_YUY2_stride;

					deinterleave_MB211_YUY2(
						src_MB, 
						dst_YUY2, 
						dst_YUY2_stride);
				}
			}
		}
	}

	if (color_format == COLORFORMAT_YV12)
	{
		if (h->config.decoder_YV12_chroma_sampling == YV12_CHROMASAMPLING_MPEG2_P)
		{
			mmx_yuy2_to_yv12(
				(uint8_t*)dst_step2,
				width*2,
				width*2, 
				(uint8_t*)dst, 
				(uint8_t*)dst + height*width + (height >> 1)*(width >> 1),
				(uint8_t*)dst + height*width,
				width,
				width >> 1,
				height);
		}
		else
		{
			mmx_yuy2_i_to_yv12(
				dst_step2,
				width*2,
				width*2,
				dst,
				dst + height*width + (height >> 1)*(width >> 1), 
				dst + height*width, 
				width,
				width >> 1,
				height);
		}
	}
	else if (color_format == COLORFORMAT_YUY2)
	{
		;
	}
	else if (color_format == COLORFORMAT_RGB32)
	{
		const int is_RGB2YUV_FULL = (h->config.decoder_YUV_to_RGB_conversion == RGB2YUV_601 ? 0 : 1);

		if (h->flip_RGB)
		{
			mmx_YUY2toRGB32(
				dst_step2 + (height-1)*width*2, 
				dst, 
				dst_step2 - width*2, 
				-width*2, 
				width*2, 
				is_RGB2YUV_FULL);
		}
		else
		{
			mmx_YUY2toRGB32(
				dst_step2, 
				dst, 
				dst_step2 + width*height*2, 
				width*2, 
				width*2, 
				is_RGB2YUV_FULL);
		}
	}
	else if (color_format == COLORFORMAT_RGB24)
	{
		const int is_RGB2YUV_FULL = (h->config.decoder_YUV_to_RGB_conversion == RGB2YUV_601 ? 0 : 1);

		if (h->flip_RGB)
		{
			mmx_YUY2toRGB24(
				dst_step2 + (height-1)*width*2, 
				dst, 
				dst_step2 - width*2, 
				-width*2, 
				width*2, 
				is_RGB2YUV_FULL);
		}
		else
		{
			mmx_YUY2toRGB24(
				dst_step2, 
				dst, 
				dst_step2 + width*height*2, 
				width*2, 
				width*2, 
				is_RGB2YUV_FULL);
		}
	}
}


static void decode_411_DV(
	void *hnd, 
	uint8_t const *src_DV, 
	uint8_t *dst)
{
	decoder_context_t *h = (decoder_context_t*)hnd;

	const int color_format = h->color_format;
	const int nsequences = h->is_system_625 ? 12 : 10;
	const int width = 720;
	const int height = h->is_system_625 ? 576 : 480;

	const int col_SB[5] = {2, 1, 3, 0, 4};
	const int row_SB[5] = {2, 6, 8, 0, 4};

	int i, j, k, p;
	int16_t * dst_5_MBs = h->buffer_5_MBs;
	uint8_t * dst_step1 = dst;
	uint8_t * dst_step2 = h->buffer_frame;

	if (color_format == COLORFORMAT_YUY2)
	{
		dst_step1 = h->buffer_frame;
	}

	for (i = 0; i < nsequences; i++)	// sequence
	{
		for (k = 0; k < 27; k++)	// MB-index in SB
		{			
			int block_indx = 5*k; 
			int DBN_1st = 7 + block_indx + block_indx / 15;	// first DIF_blocknumber
			uint8_t const * src_DIF_videosegment = src_DV + i*(150*80) + DBN_1st*(80);

			decode_DIF_videosegment(src_DIF_videosegment, dst_5_MBs, 0);
			
			for(j = 0; j < 5; j++)	// superblock row number: a,b,c,d,e
			{
				p = col_SB[j];

				// save it first as YV12 and then convert to other formats
				// todo: direct conversion to the right format
				int16_t * src_MB = dst_5_MBs + j*6*64;

				int col_MB = MB_pos_NTSC[p][2*k+0];
				int row_MB = MB_pos_NTSC[p][2*k+1] + ((row_SB[j] + i) % nsequences) * 6;

				int dst_y_pitch  = width;
				int dst_uv_pitch = width >> 2;

				uint8_t * dst_y = dst_step1 + col_MB*32 + row_MB*8*dst_y_pitch;
				uint8_t * dst_v = dst_step1 + dst_y_pitch*height + col_MB*8 + row_MB*8*dst_uv_pitch;
				uint8_t * dst_u = dst_step1 + dst_y_pitch*height + dst_uv_pitch*height + col_MB*8 + row_MB*8*dst_uv_pitch;

				if ((p == 4) && (k>23))
				{
					deinterleave_MBborder_YV12_NTSC(
						src_MB, 
						dst_y, 
						dst_u, 
						dst_v,
						dst_y_pitch, 
						dst_uv_pitch);
				}
				else
				{
					deinterleave_MB_YV12_NTSC(
						src_MB, 
						dst_y, 
						dst_u, 
						dst_v, 
						dst_y_pitch, 
						dst_uv_pitch);
				}			
			}
		}
	}

	if (color_format == COLORFORMAT_YV12)
	{
		if (h->config.decoder_YV12_chroma_sampling!=YV12_CHROMASAMPLING_DV)
		{
			YUV411_to_YUY2(
				dst_step1, 
				dst_step1 + height*width + height*(width >> 2), 
				dst_step1 + height*width,
				dst_step2, 
				width, 
				width >> 2, 
				width << 1, 
				width, 
				height);

			if (h->config.decoder_YV12_chroma_sampling == YV12_CHROMASAMPLING_MPEG2_I)
			{
				mmx_yuy2_i_to_yv12(
					dst_step2, 
					width*2, 
					width*2, 
					dst, 
					dst + height*width + height*(width >> 2), 
					dst + height*width, 
					width, 
					width >> 1,
					height);
			}
			else
			{
				mmx_yuy2_to_yv12(
					dst_step2, 
					width*2, 
					width*2, 
					dst, 
					dst + height*width + height*(width >> 2), 
					dst + height*width, 
					width, 
					width >> 1,
					height);
			}
		}
	}
	else if (color_format == COLORFORMAT_YUY2)
	{
		YUV411_to_YUY2(
			dst_step1, 
			dst_step1 + height*width + height*(width >> 2), 
			dst_step1 + height*width, 
			dst, 
			width, 
			width >> 2, 
			width << 1, 
			width, 
			height);
	}
	else if (color_format == COLORFORMAT_RGB32)
	{
		YUV411_to_YUY2(
			dst_step1, 
			dst_step1 + height*width + height*(width >> 2), 
			dst_step1 + height*width, 
			dst_step2, 
			width, 
			width >> 2, 
			width << 1, 
			width, 
			height);

		const int is_RGB2YUV_FULL = (h->config.decoder_YUV_to_RGB_conversion == RGB2YUV_601) ? 0 : 1;

		if (h->flip_RGB)
			mmx_YUY2toRGB32((uint8_t*)dst_step2+(height-1)*width*2, (uint8_t*)dst, (uint8_t*)dst_step2-width*2, -width*2, width*2, is_RGB2YUV_FULL);
		else
			mmx_YUY2toRGB32((uint8_t*)dst_step2, (uint8_t*)dst, (uint8_t*)dst_step2+width*height*2, width*2, width*2, is_RGB2YUV_FULL);
	}
	else if (color_format == COLORFORMAT_RGB24)
	{
		YUV411_to_YUY2(
			dst_step1, 
			dst_step1 + height*width + height*(width >> 2), 
			dst_step1 + height*width, 
			dst_step2, 
			width, 
			width >> 2, 
			width << 1, 
			width, 
			height);

		const int is_RGB2YUV_FULL = (h->config.decoder_YUV_to_RGB_conversion == RGB2YUV_601) ? 0 : 1;

		if (h->flip_RGB)
		{
			mmx_YUY2toRGB24(
				dst_step2 + (height-1)*width*2, 
				dst, 
				dst_step2 - width*2, 
				-width*2, 
				width*2, 
				is_RGB2YUV_FULL);
		}
		else
		{
			mmx_YUY2toRGB24(
				dst_step2, 
				dst, 
				dst_step2 + width*height*2, 
				width*2, 
				width*2, 
				is_RGB2YUV_FULL);
		}
	}
}


static void decode_DIF_videosegment(
	uint8_t const *DIF_block,
	int16_t *dst_5_MBs,
	int is_MB211)
{
	uint32_t DCT_mode[5*6];
	uint32_t quantization_number[5];
	uint32_t class_number[5*6];
	int j, l, n;
	int16_t * block = dst_5_MBs;

	decode_videosegment(
		dst_5_MBs, 
		quantization_number, 
		DCT_mode, 
		class_number, 
		DIF_block);

	n = 0;

	for (j = 0; j < 5; j++)
	{
		uint32_t QNO = quantization_number[j];

		for (l = 0; l < 6; l++)
		{
			int c1c0 = class_number[n];
			int mo = DCT_mode[n];
			int q_code = quantization_code_table_extended[(c1c0 << 4) + QNO];
			
			if (mo == 0)
			{
				int16_t const * q_step = quantization_step_88_extended + q_code * 64;

				multiply_clip_16bit(block, block, q_step);
				block[0] = block[0]*16;
				weighting_block(block, block, W_88_inv_16bit);

				idct_88(block, block);

				int k;
				for(k = 0; k < 64; k++) 
				{
					block[k] += 128;
					if ((uint16_t)(block[k]) > 255)
					{
						if (block[k]<0) 
							block[k]=0;
						else 
							block[k]=255;
					}
				}
			}
			else
			{
				int16_t const * q_step = quantization_step_248_extended + q_code * 64;

				multiply_clip_16bit(block, block, q_step);
				block[0] = block[0]*16;
				weighting_block(block, block, W_248_inv_16bit);

				idct_248(block, block);

				int k;
				for(k = 0; k < 64; k++) 
				{
					block[k] += 128;
					if ((uint16_t)(block[k]) > 255)
					{
						if (block[k]<0) 
							block[k]=0;
						else 
							block[k]=255;
					}
				}
			}

			n++;
			block += 64;

			// in case of DVCPRO50 we deal with Y0-*-Y1-*-U-V Macroblocks
			if (is_MB211 && (l == 0 || l == 2))
			{
				l += 1;
				n += 1;
				block += 64;
			}
		}
	}
}


static void deinterleave_MB_YV12_PAL(
	int16_t const *src_MB, 
	uint8_t *dst_YUV420_Y, 
	uint8_t *dst_YUV420_U, 
	uint8_t *dst_YUV420_V, 
	int dst_YUV420_Y_pitch, 
	int dst_YUV420_UV_pitch)
{
	uint8_t * x = dst_YUV420_Y;
	int16_t const * y = src_MB;
	int x_pitch = dst_YUV420_Y_pitch;
	int i,k;

	for (k=0; k<2; k++)
	{
		for(i=0; i<8; i++)
		{
			// Block_Y0
			x[0] = (uint8_t)y[i*8+0];
			x[1] = (uint8_t)y[i*8+1];
			x[2] = (uint8_t)y[i*8+2];
			x[3] = (uint8_t)y[i*8+3];
			x[4] = (uint8_t)y[i*8+4];
			x[5] = (uint8_t)y[i*8+5];
			x[6] = (uint8_t)y[i*8+6];
			x[7] = (uint8_t)y[i*8+7];

			// Block_Y1
			x[0+8] = (uint8_t)y[i*8+64+0];
			x[1+8] = (uint8_t)y[i*8+64+1];
			x[2+8] = (uint8_t)y[i*8+64+2];
			x[3+8] = (uint8_t)y[i*8+64+3];
			x[4+8] = (uint8_t)y[i*8+64+4];
			x[5+8] = (uint8_t)y[i*8+64+5];
			x[6+8] = (uint8_t)y[i*8+64+6];
			x[7+8] = (uint8_t)y[i*8+64+7];

			x=x+x_pitch;
		}

		y+=128;
	}

	x = dst_YUV420_V;
	x_pitch = dst_YUV420_UV_pitch;
	for(i=0; i<8; i++)
	{
		// Block_V
		x[0] = (uint8_t)y[i*8+0];
		x[1] = (uint8_t)y[i*8+1];
		x[2] = (uint8_t)y[i*8+2];
		x[3] = (uint8_t)y[i*8+3];
		x[4] = (uint8_t)y[i*8+4];
		x[5] = (uint8_t)y[i*8+5];
		x[6] = (uint8_t)y[i*8+6];
		x[7] = (uint8_t)y[i*8+7];

		x=x+x_pitch;
	}

	y+=64;

	x = (uint8_t*)dst_YUV420_U;
	x_pitch = dst_YUV420_UV_pitch;
	for(i=0; i<8; i++)
	{
		// Block_V
		x[0] = (uint8_t)y[i*8+0];
		x[1] = (uint8_t)y[i*8+1];
		x[2] = (uint8_t)y[i*8+2];
		x[3] = (uint8_t)y[i*8+3];
		x[4] = (uint8_t)y[i*8+4];
		x[5] = (uint8_t)y[i*8+5];
		x[6] = (uint8_t)y[i*8+6];
		x[7] = (uint8_t)y[i*8+7];

		x=x+x_pitch;
	}
}


static void deinterleave_MB211_YUY2(
	int16_t const *src_MB, 
	uint8_t *dst_YUY2, 
	int dst_YUY2_stride)
{
	uint8_t * x = dst_YUY2;
	int16_t const * y = src_MB;
	int16_t const * u = src_MB + 5*64;
	int16_t const * v = src_MB + 4*64;
	int x_stride = dst_YUY2_stride;

	int i, k;

	for (k = 0; k < 2; k++)
	{
		for (i = 0; i < 8; i++)
		{
			x[0] = (uint8_t)y[i*8+0];
			x[1] = (uint8_t)u[i*8+0];
			x[2] = (uint8_t)y[i*8+1];
			x[3] = (uint8_t)v[i*8+0];

			x[4] = (uint8_t)y[i*8+2];
			x[5] = (uint8_t)u[i*8+1];
			x[6] = (uint8_t)y[i*8+3];
			x[7] = (uint8_t)v[i*8+1];

			x[ 8] = (uint8_t)y[i*8+4];
			x[ 9] = (uint8_t)u[i*8+2];
			x[10] = (uint8_t)y[i*8+5];
			x[11] = (uint8_t)v[i*8+2];

			x[12] = (uint8_t)y[i*8+6];
			x[13] = (uint8_t)u[i*8+3];
			x[14] = (uint8_t)y[i*8+7];
			x[15] = (uint8_t)v[i*8+3];

			x += x_stride;
		}

		x += 16 - x_stride*8;
		y += 2*64;
		u += 4;
		v += 4;
	}
}


static void YUV420_to_YUY2(
	uint8_t const *src_YUV420_Y, 
	uint8_t const *src_YUV420_U, 
	uint8_t const *src_YUV420_V, 
	uint8_t *dst_YUY2, 
	int src_YUV420_Y_pitch, 
	int src_YUV420_UV_pitch, 
	int dst_YUY2_pitch, 
	int width, 
	int height)
{
	int width_half = width/2;
	int height_minus_2 = height - 2;
	int x, y;
	int UV_pitch_x2 = src_YUV420_UV_pitch << 1;

	uint8_t* out = dst_YUY2;
	uint8_t const * Y = src_YUV420_Y;
	uint8_t const * U = src_YUV420_U;
	uint8_t const * V = src_YUV420_V;

	for(x=0; x<width_half; x++)
	{
		out[0] = Y[0];
		out[1] = U[0];
		out[2] = Y[1];
		out[3] = V[0];
		out+=4;
		Y+=2;
		U++;
		V++;
	}
	out = out - 4*width_half + dst_YUY2_pitch;
	Y = Y - 2*width_half + src_YUV420_Y_pitch;
	U = U - width_half + src_YUV420_UV_pitch;
	V = V - width_half + src_YUV420_UV_pitch;

	for(x=0; x<width_half; x++)
	{
		out[0] = Y[0];
		out[1] = U[0];
		out[2] = Y[1];
		out[3] = V[0];
		out+=4;
		Y+=2;
		U++;
		V++;
	}
	out = out - 4*width_half + dst_YUY2_pitch;
	Y = Y - 2*width_half + src_YUV420_Y_pitch;
	U = U - width_half - src_YUV420_UV_pitch;
	V = V - width_half - src_YUV420_UV_pitch;


	for (y=2; y<height_minus_2; y+=4)
	{

		for(x=0; x<width_half; x++)
		{
			out[0] = Y[0];
			out[1] = U[0];
			out[2] = Y[1];
			out[3] = (uint8_t)(((int)V[0]+(int)V[UV_pitch_x2]+1)>>1);
			out+=4;
			Y+=2;
			U++;
			V++;
		}
		out = out - 4*width_half + dst_YUY2_pitch;
		Y = Y - 2*width_half + src_YUV420_Y_pitch;
		U = U - width_half + src_YUV420_UV_pitch;
		V = V - width_half + src_YUV420_UV_pitch;

		for(x=0; x<width_half; x++)
		{
			out[0] = Y[0];
			out[1] = U[0];
			out[2] = Y[1];
			out[3] = (uint8_t)(((int)V[0]+(int)V[UV_pitch_x2]+1)>>1);
			out+=4;
			Y+=2;
			U++;
			V++;
		}
		out = out - 4*width_half + dst_YUY2_pitch;
		Y = Y - 2*width_half + src_YUV420_Y_pitch;
		U = U - width_half + src_YUV420_UV_pitch;
		V = V - width_half + src_YUV420_UV_pitch;

		for(x=0; x<width_half; x++)
		{
			out[0] = Y[0];
			out[1] = (uint8_t)(((int)U[-UV_pitch_x2]+(int)U[0]+1)>>1);
			out[2] = Y[1];
			out[3] = V[0];
			out+=4;
			Y+=2;
			U++;
			V++;
		}
		out = out - 4*width_half + dst_YUY2_pitch;
		Y = Y - 2*width_half + src_YUV420_Y_pitch;
		U = U - width_half + src_YUV420_UV_pitch;
		V = V - width_half + src_YUV420_UV_pitch;

		for(x=0; x<width_half; x++)
		{
			out[0] = Y[0];
			out[1] = (uint8_t)(((int)U[-UV_pitch_x2]+(int)U[0]+1)>>1);
			out[2] = Y[1];
			out[3] = V[0];
			out+=4;
			Y+=2;
			U++;
			V++;
		}
		out = out - 4*width_half + dst_YUY2_pitch;
		Y = Y - 2*width_half + src_YUV420_Y_pitch;
		U = U - width_half - src_YUV420_UV_pitch;
		V = V - width_half - src_YUV420_UV_pitch;
	}


	for(x=0; x<width_half; x++)
	{
		out[0] = Y[0];
		out[1] = U[0];
		out[2] = Y[1];
		out[3] = V[0];
		out+=4;
		Y+=2;
		U++;
		V++;
	}
	out = out - 4*width_half + dst_YUY2_pitch;
	Y = Y - 2*width_half + src_YUV420_Y_pitch;
	U = U - width_half + src_YUV420_UV_pitch;
	V = V - width_half + src_YUV420_UV_pitch;

	for(x=0; x<width_half; x++)
	{
		out[0] = Y[0];
		out[1] = U[0];
		out[2] = Y[1];
		out[3] = V[0];
		out+=4;
		Y+=2;
		U++;
		V++;
	}

}


static void deinterleave_MB_YV12_NTSC(
	int16_t const *src_MB, 
	uint8_t *dst_y, 
	uint8_t *dst_u, 
	uint8_t *dst_v, 
	int dst_YUV420_Y_pitch, 
	int dst_YUV420_UV_pitch)
{
	int n, row;
	int stride_y = dst_YUV420_Y_pitch;
	int stride_uv = dst_YUV420_UV_pitch;

	for (n=0; n<4; n++)
	{
		for (row=0; row<8; row++)
		{
			dst_y[0] = (uint8_t)src_MB[0];
			dst_y[1] = (uint8_t)src_MB[1];
			dst_y[2] = (uint8_t)src_MB[2];
			dst_y[3] = (uint8_t)src_MB[3];
			dst_y[4] = (uint8_t)src_MB[4];
			dst_y[5] = (uint8_t)src_MB[5];
			dst_y[6] = (uint8_t)src_MB[6];
			dst_y[7] = (uint8_t)src_MB[7];
		
			src_MB += 8;
			dst_y += stride_y;
		}

		dst_y = dst_y - stride_y*8 + 8;
	}

	for (row=0; row<8; row++)
	{
		dst_v[0] = (uint8_t)src_MB[0];
		dst_v[1] = (uint8_t)src_MB[1];
		dst_v[2] = (uint8_t)src_MB[2];
		dst_v[3] = (uint8_t)src_MB[3];
		dst_v[4] = (uint8_t)src_MB[4];
		dst_v[5] = (uint8_t)src_MB[5];
		dst_v[6] = (uint8_t)src_MB[6];
		dst_v[7] = (uint8_t)src_MB[7];
	
		src_MB += 8;
		dst_v += stride_uv;
	}

	for (row=0; row<8; row++)
	{
		dst_u[0] = (uint8_t)src_MB[0];
		dst_u[1] = (uint8_t)src_MB[1];
		dst_u[2] = (uint8_t)src_MB[2];
		dst_u[3] = (uint8_t)src_MB[3];
		dst_u[4] = (uint8_t)src_MB[4];
		dst_u[5] = (uint8_t)src_MB[5];
		dst_u[6] = (uint8_t)src_MB[6];
		dst_u[7] = (uint8_t)src_MB[7];
	
		src_MB += 8;
		dst_u += stride_uv;
	}
}


static void deinterleave_MBborder_YV12_NTSC(
	int16_t const *src_MB, 
	uint8_t *dst_y, 
	uint8_t *dst_u, 
	uint8_t *dst_v, 
	int dst_YUV420_Y_pitch, 
	int dst_YUV420_UV_pitch)
{
	int n, row, col;
	int stride_y = dst_YUV420_Y_pitch;
	int stride_uv = dst_YUV420_UV_pitch;

	for (n=0; n<4; n++)
	{
		for (row=0; row<8; row++)
		{
			for(col=0; col<8; col++)
				dst_y[col] = (uint8_t)src_MB[col];
		
			src_MB += 8;
			dst_y += stride_y;
		}

		if (n == 1) 
			dst_y = dst_y - 8;
		else
			dst_y = dst_y - stride_y*8 + 8;

	}

	for (row=0; row<8; row++)
	{
		for(col=0; col<4; col++)
			dst_v[col] = (uint8_t)src_MB[col];
	
		for(col=0; col<4; col++)
			dst_v[col+8*stride_uv] = (uint8_t)src_MB[col+4];

		src_MB += 8;
		dst_v += stride_uv;
	}

	for (row=0; row<8; row++)
	{
		for(col=0; col<4; col++)
			dst_u[col] = (uint8_t)src_MB[col];
	
		for(col=0; col<4; col++)
			dst_u[col+8*stride_uv] = (uint8_t)src_MB[col+4];

		src_MB += 8;
		dst_u += stride_uv;
	}
}


static void YUV411_to_YUY2(
	uint8_t const *src_YUV420_Y, 
	uint8_t const *src_YUV420_U, 
	uint8_t const *src_YUV420_V, 
	uint8_t *dst_YUY2, 
	int src_YUV420_Y_pitch, 
	int src_YUV420_UV_pitch, 
	int dst_YUY2_pitch, 
	int width, 
	int height)
{
	int row, col;
	int width_quarter = width/4;
	uint8_t const *src_y = src_YUV420_Y;
	uint8_t const *src_u = src_YUV420_U;
	uint8_t const *src_v = src_YUV420_V;
	uint8_t *dst = dst_YUY2;

	for (row=0; row<height; row++)
	{
		for(col=0; col<(width_quarter-1); col++)
		{
			dst[col*8+0] = src_y[col*4+0];
			dst[col*8+1] = src_u[col  +0];
			dst[col*8+2] = src_y[col*4+1];
			dst[col*8+3] = src_v[col  +0];
			dst[col*8+4] = src_y[col*4+2];
			dst[col*8+5] = (uint8_t)(((int)src_u[col  +0] + (int)src_u[col  +1] + 1)>>1);
			dst[col*8+6] = src_y[col*4+3];
			dst[col*8+7] = (uint8_t)(((int)src_v[col  +0] + (int)src_v[col  +1] + 1)>>1);
		}
		dst[col*8+0] = src_y[col*4+0];
		dst[col*8+1] = src_u[col  +0];
		dst[col*8+2] = src_y[col*4+1];
		dst[col*8+3] = src_v[col  +0];
		dst[col*8+4] = src_y[col*4+2];
		dst[col*8+5] = src_u[col  +0];
		dst[col*8+6] = src_y[col*4+3];
		dst[col*8+7] = src_v[col  +0];

		dst += dst_YUY2_pitch;

		src_y += src_YUV420_Y_pitch;
		src_u += src_YUV420_UV_pitch;
		src_v += src_YUV420_UV_pitch;
	}
}
