/******************************************************************************/
/*                                                                            */
/*  Cedocida DV-Codec                                                         */
/*                                                                            */
/*  Copyright (C) 2005 Andreas Dittrich <dittrich@cithraidt.de>               */
/*  Project start: 15.12.2003                                                 */
/*  Last change  : 15.06.2007                                                 */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/*  This program is free software; you can redistribute it and/or modify      */
/*  it under the terms of the GNU General Public License as published by      */
/*  the Free Software Foundation; either version 2 of the License, or         */
/*  (at your option) any later version.                                       */
/*                                                                            */
/*  This program is distributed in the hope that it will be useful,           */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             */
/*  GNU General Public License for more details.                              */
/*                                                                            */
/*  You should have received a copy of the GNU General Public License         */
/*  along with this program; if not, write to the Free Software               */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
/*                                                                            */
/******************************************************************************/

#include <stdlib.h>
#include <string.h>

#include "inttypes.h"

#include "dv_spec.h"
#include "dct/dct.h"
#include "vlc/encode_vlc.h"
#include "signalproc.h"
#include "encode_tables.h"
#include "config.h"
#include "encode.h"

// todo: do direct inplace conversion 
#include "convert/convert.h"


#define VIDEOSYSTEM_NTSC	0
#define VIDEOSYSTEM_PAL		1


typedef struct {
	uint8_t *buffer_base;
	int16_t *buffer_5_QTMBs;
	int16_t *buffer_5_MBs;
	uint8_t *buffer_frame;
	int16_t *buffer_block;

	int color_format;
	int flip_RGB;
	int is_system_625;
	dvformat_t dvfmt;

	int nbytes_dv;
	int nbytes_frame;

	configvars_t config;
} encoder_context_t;



static void encode_420_DV(void *hnd, uint8_t const *src, uint8_t *dst_DV);
static void encode_422_DV(void *hnd, uint8_t const *src, uint8_t *dst_DV);
static void encode_411_DV(void *hnd, uint8_t const *src, uint8_t *dst_DV);

static void encode_6_Header_DIF_blocks(
	uint8_t *DIF_block, 
	int DIF_sequence_number, 
	int video_system,
	dvformat_t dvfmt,
	int is_16to9,
	int FSC);

static void encode_Audio_DIF_block(
	uint8_t *DIF_block, 
	int DIF_sequence_number, 
	int DIF_block_number, 
	int video_system,
	dvformat_t dvfmt,
	int ch);

static void encode_5_Video_DIF_blocks(
	encoder_context_t * h, 
	int16_t const *src_5_MBs, 
	uint8_t *dst_DIF_block, 
	int sequence_number, 
	int block_number,
	int FSC,
	int is_MB211);



#define LOG2_ALIGN           7
#define SIZE_BUFFER_5_MBS    (5*6*64*2)
#define SIZE_BUFFER_5_QTMBS  (5*6*64*2)
#define SIZE_BUFFER_FRAME    (720*576*4)
#define SIZE_BUFFER_BLOCK    (64*2)


void * encoder_open()
{
	encoder_context_t * const h = (encoder_context_t*)malloc(sizeof(encoder_context_t));
	if (h == NULL)
		return NULL;

	memset(h, 0x00, sizeof(*h));

	h->buffer_base = (uint8_t*)malloc(((1 << LOG2_ALIGN) - 1) + 
		SIZE_BUFFER_FRAME + 
		SIZE_BUFFER_5_QTMBS +
		SIZE_BUFFER_5_MBS +
		SIZE_BUFFER_BLOCK);

	if (h->buffer_base == NULL)
		encoder_close(h);

	h->buffer_frame    = h->buffer_base + ((int)((uint8_t*)NULL - h->buffer_base) & ((1 << LOG2_ALIGN) - 1));
	h->buffer_5_QTMBs  = (int16_t*)(h->buffer_frame + SIZE_BUFFER_FRAME);
	h->buffer_5_MBs    = (int16_t*)((uint8_t*)(h->buffer_5_QTMBs) + SIZE_BUFFER_5_QTMBS);
	h->buffer_block    = (int16_t*)((uint8_t*)(h->buffer_5_MBs) + SIZE_BUFFER_5_MBS);

	h->color_format = COLORFORMAT_YV12;

	return h;
}


void encoder_configure(
	void *hnd, 
	int color_format, 
	int flip_RGB, 
	int is_system_625,
	dvformat_t dvfmt,
	configvars_t const *pconfig)
{
	encoder_context_t * const h = (encoder_context_t*)hnd;

	if (h == NULL)
		return;

	h->color_format = color_format;
	h->flip_RGB = flip_RGB;
	h->is_system_625 = is_system_625;
	h->dvfmt = dvfmt;

	int const nsequences = is_system_625 ? 12 : 10;
	int const nchannel = (dvfmt == DVFORMAT_DV50) ? 2 : 1;
	int const width = 720;
	int const height = is_system_625 ? 576 : 480;

	h->nbytes_dv = nchannel * nsequences * 150 * 80;

	h->nbytes_frame = height *
		(h->color_format == COLORFORMAT_YV12)  ? width * height * 3 / 2 :
		(h->color_format == COLORFORMAT_YUY2)  ? width * height * 2 :
		(h->color_format == COLORFORMAT_RGB24) ? (width * height * 3 + 3) & ~0x03  :
		(h->color_format == COLORFORMAT_RGB32) ? width * height * 4 :
		width * height * 4;

	memcpy(&h->config, pconfig, sizeof(configvars_t));
}


void encoder_close(void *hnd)
{
	encoder_context_t * const h = (encoder_context_t*)hnd;

	if (h == NULL)
		return;

	if (h->buffer_base != NULL)
		free(h->buffer_base);

	free(h);
}


int encode_DV(
	void *hnd, 
	uint8_t const *src, 
	unsigned int size_src, 
	uint8_t *dst_DV, 
	unsigned int *p_size_dst)
{
	encoder_context_t * const h = (encoder_context_t*)hnd;

	if (h == NULL || p_size_dst == NULL)
		return -1;

	if ((int)*p_size_dst < h->nbytes_dv ||
		(int)size_src < h->nbytes_frame)
	{
		return -2;
	}

	switch (h->dvfmt) 
	{
	case DVFORMAT_DVSD:
	{
		if (h->is_system_625)
		{
			encode_420_DV(h, src, dst_DV);
		}
		else
		{
			encode_411_DV(h, src, dst_DV);
		}

		break;
	}

	case DVFORMAT_DV25:
		encode_411_DV(h, src, dst_DV);
		break;

	case DVFORMAT_DV50:
		encode_422_DV(h, src, dst_DV);
		break;

	default:
		; // do nothing
	}

	emms_asm();
	
	return 0;
}


static void encode_420_DV(void *hnd, uint8_t const *src, uint8_t *dst_DV)
{
	encoder_context_t * const h = (encoder_context_t*)hnd;

	int color_format = h->color_format;
	const int width  = 720;
	const int height = 576;

	//const int row_SB_offset[5] = { 2, 6, 8, 0, 4};
	static const int col_SB[5] = { 2, 1, 3, 0, 4};
	static const int row_SB[12*5] = { 
		2, 3, 4, 5, 6, 7, 8, 9,10,11, 0, 1,
		6, 7, 8, 9,10,11, 0, 1, 2, 3, 4, 5,
		8, 9,10,11, 0, 1, 2, 3, 4, 5, 6, 7,
		0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,
		4, 5, 6, 7, 8, 9,10,11, 0, 1, 2, 3
	};

	int i, r, j;

	int16_t * const dst_MB = h->buffer_5_MBs;
	int16_t * const dst_MB_a = dst_MB + 0*6*64;
	int16_t * const dst_MB_b = dst_MB + 1*6*64;
	int16_t * const dst_MB_c = dst_MB + 2*6*64;
	int16_t * const dst_MB_d = dst_MB + 3*6*64;
	int16_t * const dst_MB_e = dst_MB + 4*6*64;

	// quick and dirty color-conversion (should be better done in combination with later "block-interleaving")
	if (color_format == COLORFORMAT_YV12) 
	{
		if (h->config.encoder_YV12_chroma_sampling == YV12_CHROMASAMPLING_MPEG2_P)
		{
			mmx_yv12_to_yuy2(
				src,
				src + height*width + (height >> 1)*(width >> 1),
				src + height*width,
				width,
				width,
				width >> 1,
				h->buffer_frame,
				width*2,
				height);

			color_format = COLORFORMAT_YUY2;
			src = h->buffer_frame;
		}
		else if (h->config.encoder_YV12_chroma_sampling == YV12_CHROMASAMPLING_MPEG2_I)
		{
			mmx_yv12_i_to_yuy2(
				src,
				src + height*width + (height >> 1)*(width >> 1), 
				src + height*width, 
				width,
				width,
				width >> 1,
				h->buffer_frame,
				width*2,
				height);

			color_format = COLORFORMAT_YUY2;
			src = h->buffer_frame;
		}
	}
	else if (color_format == COLORFORMAT_RGB32)
	{
		if (h->flip_RGB)
		{
			mmx_ConvertRGB32toYUY2_ad(
				src + (height-1)*width*4, 
				h->buffer_frame,
				-width*4,
				width << 1,
				width,
				height,
				h->config.encoder_RGB_to_YUV_conversion);
		}
		else
		{
			mmx_ConvertRGB32toYUY2_ad(
				src, 
				h->buffer_frame,
				width*4, 
				width << 1, 
				width, 
				height, 
				h->config.encoder_RGB_to_YUV_conversion);
		}

		color_format = COLORFORMAT_YUY2;
		src = h->buffer_frame;
	}
	else if (color_format == COLORFORMAT_RGB24)
	{
		if (h->flip_RGB)
		{
			mmx_ConvertRGB24toYUY2_ad(
				src + ((height-1)*width*3),
				h->buffer_frame,
				-width*3,
				width << 1,
				width,
				height,
				h->config.encoder_RGB_to_YUV_conversion);
		}
		else
		{
			mmx_ConvertRGB24toYUY2_ad(
				src,
				h->buffer_frame,
				width*3,
				width << 1,
				width,
				height,
				h->config.encoder_RGB_to_YUV_conversion);
		}

		color_format = COLORFORMAT_YUY2;
		src = h->buffer_frame;
	}
	else if (color_format != COLORFORMAT_YUY2)
	{
		return; //should never happen!
	}

	for (i = 0; i < 12; i++)	// sequence
	{
		encode_6_Header_DIF_blocks(dst_DV, i, VIDEOSYSTEM_PAL, h->dvfmt, h->config.encoder_aspect_ratio, 0);

		for (r = 0; r < 3; r++)	// row_MB in SB
		{
			if (color_format == COLORFORMAT_YV12)
			{
				uint8_t const * src_y = src;
				uint8_t const * src_v = src_y + width * height;
				uint8_t const * src_u = src_v + (width >> 1) * (height >> 1);

				int delta_y_a = col_SB[0]*9*16 + (row_SB[0*12+i]*3+r)*(16*width);
				int delta_y_b = col_SB[1]*9*16 + (row_SB[1*12+i]*3+r)*(16*width);
				int delta_y_c = col_SB[2]*9*16 + (row_SB[2*12+i]*3+r)*(16*width);
				int delta_y_d = col_SB[3]*9*16 + (row_SB[3*12+i]*3+r)*(16*width);
				int delta_y_e = col_SB[4]*9*16 + (row_SB[4*12+i]*3+r)*(16*width);

				int delta_uv_a = col_SB[0]*9*8 + (row_SB[0*12+i]*3+r)*(8*width/2);
				int delta_uv_b = col_SB[1]*9*8 + (row_SB[1*12+i]*3+r)*(8*width/2);
				int delta_uv_c = col_SB[2]*9*8 + (row_SB[2*12+i]*3+r)*(8*width/2);
				int delta_uv_d = col_SB[3]*9*8 + (row_SB[3*12+i]*3+r)*(8*width/2);
				int delta_uv_e = col_SB[4]*9*8 + (row_SB[4*12+i]*3+r)*(8*width/2);

				for (j = 0; j < 9; j++)	// col_MB in SB
				{
					int DBN_video = 5 * SB_row_order[r][j]; //DIF_blocknumber of video
					int DBN_audio = DBN_video / 15;	//DIF_blocknumber of audio
					int block_index = 7 + DBN_video + DBN_audio;
					uint8_t * dst_5_DIF_blocks = dst_DV + block_index * 80;

					if ((DBN_video % 15) == 0)
					{
						encode_Audio_DIF_block(dst_5_DIF_blocks - 80, i, DBN_audio, VIDEOSYSTEM_PAL, h->dvfmt, 0);
					}

					const int stride_uv = width >> 1;

					copy_YV12_to_MBYVU420_12bit(src_y + delta_y_a, src_u + delta_uv_a, src_v + delta_uv_a, dst_MB_a, width, stride_uv);
					copy_YV12_to_MBYVU420_12bit(src_y + delta_y_b, src_u + delta_uv_b, src_v + delta_uv_b, dst_MB_b, width, stride_uv);
					copy_YV12_to_MBYVU420_12bit(src_y + delta_y_c, src_u + delta_uv_c, src_v + delta_uv_c, dst_MB_c, width, stride_uv);
					copy_YV12_to_MBYVU420_12bit(src_y + delta_y_d, src_u + delta_uv_d, src_v + delta_uv_d, dst_MB_d, width, stride_uv);
					copy_YV12_to_MBYVU420_12bit(src_y + delta_y_e, src_u + delta_uv_e, src_v + delta_uv_e, dst_MB_e, width, stride_uv);
					src_y += 16;
					src_u += 8;
					src_v += 8;

					encode_5_Video_DIF_blocks(h, dst_MB, dst_5_DIF_blocks, i, DBN_video, 0, 0);
				}
			}
			else //YUY2
			{
				uint8_t const * src_yuy2 = src;

				int delta_y_a = col_SB[0]*9*(16*2) + (row_SB[0*12+i]*3+r)*(16*width*2);
				int delta_y_b = col_SB[1]*9*(16*2) + (row_SB[1*12+i]*3+r)*(16*width*2);
				int delta_y_c = col_SB[2]*9*(16*2) + (row_SB[2*12+i]*3+r)*(16*width*2);
				int delta_y_d = col_SB[3]*9*(16*2) + (row_SB[3*12+i]*3+r)*(16*width*2);
				int delta_y_e = col_SB[4]*9*(16*2) + (row_SB[4*12+i]*3+r)*(16*width*2);

				for (j = 0; j < 9; j++)	// col_MB in SB
				{
					int DBN_video = 5 * SB_row_order[r][j]; //DIF_blocknumber of video
					int DBN_audio = DBN_video / 15;	//DIF_blocknumber of audio
					int block_index = 7 + DBN_video + DBN_audio;
					uint8_t * dst_5_DIF_blocks = dst_DV + block_index * 80;

					if ((DBN_video % 15) == 0)
					{
						encode_Audio_DIF_block(dst_5_DIF_blocks - 80, i, DBN_audio, VIDEOSYSTEM_PAL, h->dvfmt, 0);
					}

					const int stride_yuy2 = width << 1;

					copy_YUY2_to_MBYVU420_12bit(src_yuy2 + delta_y_a, dst_MB_a, stride_yuy2);
					copy_YUY2_to_MBYVU420_12bit(src_yuy2 + delta_y_b, dst_MB_b, stride_yuy2);
					copy_YUY2_to_MBYVU420_12bit(src_yuy2 + delta_y_c, dst_MB_c, stride_yuy2);
					copy_YUY2_to_MBYVU420_12bit(src_yuy2 + delta_y_d, dst_MB_d, stride_yuy2);
					copy_YUY2_to_MBYVU420_12bit(src_yuy2 + delta_y_e, dst_MB_e, stride_yuy2);
					src_yuy2 += 16*2;

					encode_5_Video_DIF_blocks(h, dst_MB, dst_5_DIF_blocks, i, DBN_video, 0, 0);
				}
			}
		}

		dst_DV += 150*80;
	}
}


static void encode_422_DV(void *hnd, uint8_t const *src, uint8_t *dst_DV)
{
	encoder_context_t * const h = (encoder_context_t*)hnd;

	int color_format = h->color_format;
	const int width  = 720;
	const int height = h->is_system_625 ? 576 : 480;
	const int nsequences = h->is_system_625 ? 12 : 10;

	static const int col_SB[5] = { 2, 1, 3, 0, 4 };
	static const int row_SB[5] = { 2, 6, 8, 0, 4 };

	int i, r, j, ch;

	int16_t * const dst_MB = h->buffer_5_MBs;
	int16_t * const dst_MB_a = dst_MB + 0*6*64;
	int16_t * const dst_MB_b = dst_MB + 1*6*64;
	int16_t * const dst_MB_c = dst_MB + 2*6*64;
	int16_t * const dst_MB_d = dst_MB + 3*6*64;
	int16_t * const dst_MB_e = dst_MB + 4*6*64;

	if (color_format == COLORFORMAT_YV12) 
	{
		if (h->config.encoder_YV12_chroma_sampling == YV12_CHROMASAMPLING_MPEG2_P)
		{
			mmx_yv12_to_yuy2(
				src,
				src + height*width + (height >> 1)*(width >> 1),
				src + height*width,
				width,
				width,
				width >> 1,
				h->buffer_frame,
				width*2,
				height);

			src = h->buffer_frame;
		}
		else
		{
			mmx_yv12_i_to_yuy2(
				src,
				src + height*width + (height >> 1)*(width >> 1), 
				src + height*width, 
				width,
				width,
				width >> 1,
				h->buffer_frame,
				width*2,
				height);

			src = h->buffer_frame;
		}
	}
	else if (color_format == COLORFORMAT_RGB32)
	{
		if (h->flip_RGB)
		{
			mmx_ConvertRGB32toYUY2_ad(
				src + (height-1)*width*4, 
				h->buffer_frame,
				-width*4,
				width << 1,
				width,
				height,
				h->config.encoder_RGB_to_YUV_conversion);
		}
		else
		{
			mmx_ConvertRGB32toYUY2_ad(
				src, 
				h->buffer_frame,
				width*4, 
				width << 1, 
				width, 
				height, 
				h->config.encoder_RGB_to_YUV_conversion);
		}

		src = h->buffer_frame;
	}
	else if (color_format == COLORFORMAT_RGB24)
	{
		if (h->flip_RGB)
		{
			mmx_ConvertRGB24toYUY2_ad(
				src + ((height-1)*width*3),
				h->buffer_frame,
				-width*3,
				width << 1,
				width,
				height,
				h->config.encoder_RGB_to_YUV_conversion);
		}
		else
		{
			mmx_ConvertRGB24toYUY2_ad(
				src,
				h->buffer_frame,
				width*3,
				width << 1,
				width,
				height,
				h->config.encoder_RGB_to_YUV_conversion);
		}

		src = h->buffer_frame;
	}
	else if (color_format != COLORFORMAT_YUY2)
	{
		return; //should never happen!
	}

	for (ch = 0; ch < 2; ch++)
	for (i = 0; i < nsequences; i++)	// sequence
	{
		encode_6_Header_DIF_blocks(
			dst_DV, 
			i, 
			h->is_system_625 ? VIDEOSYSTEM_PAL : VIDEOSYSTEM_NTSC, 
			h->dvfmt, 
			h->config.encoder_aspect_ratio,
			ch);

		for (r = 0; r < 3; r++)	// row_MB in SB
		{
			uint8_t const * src_yuy2 = src;
			int const stride_yuy2 = width << 1;

			int delta_y_a = col_SB[0]*9*(16*2) + ((((row_SB[0] + i) % nsequences)*2+ch)*3+r)*8*stride_yuy2;
			int delta_y_b = col_SB[1]*9*(16*2) + ((((row_SB[1] + i) % nsequences)*2+ch)*3+r)*8*stride_yuy2;
			int delta_y_c = col_SB[2]*9*(16*2) + ((((row_SB[2] + i) % nsequences)*2+ch)*3+r)*8*stride_yuy2;
			int delta_y_d = col_SB[3]*9*(16*2) + ((((row_SB[3] + i) % nsequences)*2+ch)*3+r)*8*stride_yuy2;
			int delta_y_e = col_SB[4]*9*(16*2) + ((((row_SB[4] + i) % nsequences)*2+ch)*3+r)*8*stride_yuy2;

			for (j = 0; j < 9; j++)	// col_MB in SB
			{
				uint8_t DBN_video = 5*SB_row_order[r][j]; //DIF_blocknumber of video
				uint8_t DBN_audio = DBN_video / 15;	//DIF_blocknumber of audio
				int block_index = 7 + DBN_video + DBN_audio;
				uint8_t * dst_5_DIF_blocks = dst_DV + block_index * 80;

				if ((DBN_video % 15) == 0)
				{
					encode_Audio_DIF_block(dst_5_DIF_blocks - 80, i, DBN_audio, VIDEOSYSTEM_PAL, h->dvfmt, ch);
				}

				copy_YUY2_to_MBYVU211_12bit_c(src_yuy2 + delta_y_a, dst_MB_a, stride_yuy2);
				copy_YUY2_to_MBYVU211_12bit_c(src_yuy2 + delta_y_b, dst_MB_b, stride_yuy2);
				copy_YUY2_to_MBYVU211_12bit_c(src_yuy2 + delta_y_c, dst_MB_c, stride_yuy2);
				copy_YUY2_to_MBYVU211_12bit_c(src_yuy2 + delta_y_d, dst_MB_d, stride_yuy2);
				copy_YUY2_to_MBYVU211_12bit_c(src_yuy2 + delta_y_e, dst_MB_e, stride_yuy2);
				src_yuy2 += 16*2;

				encode_5_Video_DIF_blocks(h, dst_MB, dst_5_DIF_blocks, (uint8_t)i, DBN_video, ch, 1);
			}
		}

		dst_DV += 150 * 80;
	}
}


static void encode_411_DV(void *hnd, uint8_t const *src, uint8_t *dst_DV)
{
	encoder_context_t * const h = (encoder_context_t*)hnd;

	const int nsequences = h->is_system_625 ? 12 : 10;
	const char video_system = h->is_system_625 ? VIDEOSYSTEM_PAL : VIDEOSYSTEM_NTSC;
	const int height = h->is_system_625 ? 576 : 480;
	const int width = 720;

	static const int col_SB[5] = {2, 1, 3, 0, 4};
	static const int row_SB[5] = {2, 6, 8, 0, 4}; 

	int color_format = h->color_format;
	int i, k, j, p;

	uint8_t const *src_yuy2;
	uint8_t const *src_y;
	uint8_t const *src_u;
	uint8_t const *src_v;

	int16_t * dst_MB = h->buffer_5_MBs;

	// quick and dirty color-conversion (should be better done in combination with later "block-interleaving")
	if (color_format == COLORFORMAT_YV12) 
	{
		if (h->config.encoder_YV12_chroma_sampling == YV12_CHROMASAMPLING_MPEG2_P)
		{
			mmx_yv12_to_yuy2(
				src,
				src + height*width + height*width/4,
				src + height*width,
				width, 
				width, 
				width >> 1,
				h->buffer_frame,
				width * 2,
				height);

			color_format = COLORFORMAT_YUY2;
			src = h->buffer_frame;
		}
		else if (h->config.encoder_YV12_chroma_sampling == YV12_CHROMASAMPLING_MPEG2_I)
		{
			mmx_yv12_i_to_yuy2(
				src,
				src + height*width + height*width/4,
				src + height*width,
				width,
				width,
				width >> 1,
				h->buffer_frame,
				width*2,
				height);

			color_format = COLORFORMAT_YUY2;
			src = h->buffer_frame;
		}
	}
	else if (color_format == COLORFORMAT_RGB32)
	{
		if (h->flip_RGB)
		{
			mmx_ConvertRGB32toYUY2_ad(
				src + (height-1)*width*4,
				h->buffer_frame,
				-width*4,
				width << 1,
				width,
				height,
				h->config.encoder_RGB_to_YUV_conversion);
		}
		else
		{
			mmx_ConvertRGB32toYUY2_ad(
				src,
				h->buffer_frame, 
				width*4, 
				width << 1, 
				width, 
				height, 
				h->config.encoder_RGB_to_YUV_conversion);
		}

		color_format = COLORFORMAT_YUY2;
		src = h->buffer_frame;
	}
	else if (color_format == COLORFORMAT_RGB24)
	{
		if (h->flip_RGB)
		{
			mmx_ConvertRGB24toYUY2_ad(
				src + (height-1)*width*3,
				h->buffer_frame,
				-width*3, 
				width << 1, 
				width, 
				height, 
				h->config.encoder_RGB_to_YUV_conversion);
		}
		else
		{
			mmx_ConvertRGB24toYUY2_ad(
				src,
				h->buffer_frame,
				width*3, 
				width << 1, 
				width, 
				height, 
				h->config.encoder_RGB_to_YUV_conversion);
		}

		color_format = COLORFORMAT_YUY2;
		src = h->buffer_frame;
	}
	else if (color_format != COLORFORMAT_YUY2)
	{
		return; //should never happen!
	}

	for (i = 0; i < nsequences; i++)	// sequence
	{
		encode_6_Header_DIF_blocks(dst_DV, i, video_system, h->dvfmt, h->config.encoder_aspect_ratio, 0);

		for (k = 0; k < 27; k++)	// MB-index in SB
		{
			uint8_t DBN_video = (uint8_t)(5*k); //DIF_blocknumber of video
			uint8_t DBN_audio = DBN_video / (uint8_t)15;	//DIF_blocknumber of audio
			int block_index = 7 + DBN_video + DBN_audio;
			uint8_t * dst_5_DIF_blocks = dst_DV + block_index*80;

			if ((DBN_video % 15) == 0)
			{
				encode_Audio_DIF_block(dst_5_DIF_blocks - 80, i, DBN_audio, video_system, h->dvfmt, 0);
			}

			for(j = 0; j < 5; j++)	// superblock row number: a,b,c,d,e
			{
				p = col_SB[j];
				int16_t * dst_MB_j = dst_MB + j*6*64;
				int col_MB = MB_pos_NTSC[p][2*k+0];
				int row_MB = MB_pos_NTSC[p][2*k+1] + ((row_SB[j] + i) % nsequences) * 6;

				if (color_format == COLORFORMAT_YV12)	// YV12
				{
					int const stride_y = width;
					int const stride_uv = width >> 2;

					src_y = src + col_MB*32 + row_MB*8*stride_y;
					src_v = src + stride_y*height + col_MB*8 + row_MB*8*stride_uv;
					src_u = src_v + stride_uv*height;

					if ((p == 4) && (k > 23))
						copy_YV12_to_MBYVU411border_12bit(src_y, src_u, src_v, dst_MB_j, stride_y, stride_uv);
					else
						copy_YV12_to_MBYVU411_12bit(src_y, src_u, src_v, dst_MB_j, stride_y, stride_uv);
				}
				else if (color_format == COLORFORMAT_YUY2) //YUY2
				{
					src_yuy2 = src + col_MB*32*2 + row_MB*8*width*2;

					if ((p == 4) && (k > 23))
						copy_YUY2_to_MBYVU411border_12bit(src_yuy2, dst_MB_j, width*2);
					else
						copy_YUY2_to_MBYVU411_12bit(src_yuy2, dst_MB_j, width*2);
				}
			}

			encode_5_Video_DIF_blocks(h, dst_MB, dst_5_DIF_blocks, (uint8_t)i, DBN_video, 0, 0);
		}

		dst_DV += 150*80;
	}
}


static void encode_6_Header_DIF_blocks(
	uint8_t *DIF_block, 
	int DIF_sequence_number, 
	int video_system, 
	dvformat_t dvfmt,
	int is_16to9,
	int ch)
{
	const uint8_t SCT_head = 0;     // DIF-block-type: head
	const uint8_t SCT_subcode = 1;  // DIF-block-type: subcode
	const uint8_t SCT_vaux = 2;     // DIF-block-type: vaux
	const uint8_t Seq = 15;         // video-sequencenumber: n/a
	const uint8_t Dseq = (uint8_t)DIF_sequence_number;
	const uint8_t FSC = (uint8_t)ch;
	const uint8_t vsys = (uint8_t)video_system;
	
	uint8_t bcsys = (video_system == VIDEOSYSTEM_PAL && dvfmt == DVFORMAT_DVSD && is_16to9) ? 1 : 0;
	uint8_t disp = 
		(!is_16to9) ? 0 :
		(video_system == VIDEOSYSTEM_PAL && dvfmt == DVFORMAT_DVSD) ? 0x07 : 0x02;
	
	// set all bytes to 0xFF
	ones_block80(DIF_block, 6);	

	uint8_t (*DB)[80] = (uint8_t(*)[80])DIF_block;

	// head
	DB[0][0] = (SCT_head << 5) | 0x10 | Seq;
	DB[0][1] = (Dseq << 4) | (FSC << 3) | 0x07;
	DB[0][2] = 0;

	const uint8_t APX = (dvfmt == DVFORMAT_DV25 || dvfmt == DVFORMAT_DV50) ? 0x01 : 0x00; // APT, AP0, AP1, AP2

	uint8_t * PC = &DB[0][3];	// head data
	PC[0] = 0x3F | (vsys << 7);
	PC[1] = 0xF8 | APX;	//  DFTIA=1111: n/a, 1, APT
	PC[2] = 0x78 | APX;	// 
	PC[3] = 0x78 | APX;	//  
	PC[4] = 0x78 | APX;	// 

	// subcode 0, 1
	DB[1][0] = (SCT_subcode << 5) | 0x10 | Seq;
	DB[1][1] = (Dseq << 4) | (FSC << 3) | 0x07;
	DB[1][2] = 0;

	DB[2][0] = (SCT_subcode << 5) | 0x10 | Seq;
	DB[2][1] = (Dseq << 4) | (FSC << 3) | 0x07;
	DB[2][2] = 1;

	// vaux 0, 1, 2
	DB[3][0] = (SCT_vaux << 5) | 0x10 | Seq;
	DB[3][1] = (Dseq << 4) | (FSC << 3) | 0x07;
	DB[3][2] = 0;

	DB[4][0] = (SCT_vaux << 5) | 0x10 | Seq;
	DB[4][1] = (Dseq << 4) | (FSC << 3) | 0x07;
	DB[4][2] = 1;

	DB[5][0] = (SCT_vaux << 5) | 0x10 | Seq;
	DB[5][1] = (Dseq << 4) | (FSC << 3) | 0x07;
	DB[5][2] = 2;

	if (DIF_sequence_number & 0x01)	// even or odd? see: IEC 61834-2 / table 32
		PC = &DB[3][3];			// video-packet 0
	else
		PC = &DB[5][3] + 9*5;	// video-packet 39

	uint8_t const stype = (dvfmt == DVFORMAT_DV50) ? 0x04 : 0;	// 4:2:2 or 4:1:1 compression (in case of DVCPRO, 0 else)
	uint8_t const visc  = ((dvfmt == DVFORMAT_DV25) || (dvfmt == DVFORMAT_DV50)) ? 0x7F : 0xFF;

	//VS
	PC[0] = 0x60;	// source
	PC[1] = 0xFF;	// no information
	PC[2] = 0xFF;	// CLF not valid
	PC[3] = 0xC0 | (vsys << 5) | stype;	// system 625-50 (PAL)
	PC[4] = visc;

	//VSC
	PC += 5;
	PC[0] = 0x61;	// source control
	PC[1] = 0x3F;	// CGMS=00: copy allowed, ISR=11: n/a, CMP=11: n/a, SS=11: n/a
	PC[2] = 0xC8 | disp;	// REC ST=1: no startpoint, 1, REC MODE=00: original, 1, disp
	PC[3] = 0xFC | bcsys;	// FF=1: both fields, FS=1: field1 first, FC=1: different frame, IL=1: interlaced, ST=1: no picture, SC=1: no picture, BCSYS
	PC[4] = 0xFF;	// 1, category=1111111

	// next optional packets are date, time ...
}


static void encode_Audio_DIF_block(
	uint8_t *DIF_block, 
	int DIF_sequence_number, 
	int DIF_block_number, 
	int video_system,
	dvformat_t dvfmt,
	int ch)
{
	const uint8_t SCT_audio = 3;	// DIF-block-type: audio
	const uint8_t Seq = 15;	// Video-sequenznumber: not available
	const uint8_t Dseq = (uint8_t)DIF_sequence_number;
	const uint8_t DBN = (uint8_t)DIF_block_number;
	const uint8_t FSC = (uint8_t)ch;
	const uint8_t vsys = (uint8_t)video_system;

	const uint8_t AF_SIZE = 
		(dvfmt == DVFORMAT_DVSD) ? 0 : 
		(video_system == VIDEOSYSTEM_PAL) ? 24 : 22;

	const uint8_t PA  = (dvfmt == DVFORMAT_DVSD) ? 0 : 1;
	const uint8_t EFC = (dvfmt == DVFORMAT_DVSD) ? 3 : 0;

	// set all bytes to zero
	clear_block80(DIF_block, 1);	

	DIF_block[0] = (SCT_audio << 5) | 0x10 | Seq;
	DIF_block[1] = (Dseq << 4) | (FSC << 3) | 0x07;
	DIF_block[2] = DBN;

	uint8_t * PC = DIF_block + 3;	// AAUX packet

	if ((DIF_sequence_number & 0x01) == 0)	// even?	see: IEC 61834-2 / table 31
		DIF_block_number = DIF_block_number - 3;	// only to combine the next if-section

	if (DIF_block_number == 0)
	{
		PC[0] = 0x50;	// source 
		PC[1] = 0x40 | AF_SIZE;	// LF=0: locked mode, 1, AF-size
		PC[2] = 0x0F | (PA << 4);	// SM=0, CHN=00, PA=0, audio-mode=1111: n/a, 
		PC[3] = 0xC0 | (vsys << 5);	// 1, ML=1, system=x, stype=00000
		PC[4] = 0xC0;	// EF=1, TC=1, SMP=000: 48kHz, QU=000: 16bit
	}
	else if (DIF_block_number == 1)
	{
		PC[0] = 0x51;	// source control
		PC[1] = 0x3C | EFC;	// CGMS=00: copy allowed, ISR=11: n/a, CMP=11: n/a, SS=11: n/a
		PC[2] = 0xFF;	// REC ST=1: no startpoint, REC END=1, REC MODE=111: not valid, Insert-CH=111: n/a
		PC[3] = 0xFF;	// DRF=1: forward, speed=1111111: n/a
		PC[4] = 0xFF;	// 1, category=1111111
	}
	else	// next 4 optional packets are e.g. date, ...
	{
		PC[0] = 0xFF;	// NOI-packet 
		PC[1] = 0xFF;	//  
		PC[2] = 0xFF;	//  
		PC[3] = 0xFF;	//  
		PC[4] = 0xFF;	// 
	}
}


static void encode_5_Video_DIF_blocks(
	encoder_context_t* h, 
	int16_t const *src_5_MBs, 
	uint8_t *dst_DIF_block, 
	int sequence_number, 
	int block_number,
	int ch,
	int is_MB211)
{
	int16_t * const buffer_block = h->buffer_block;
	int16_t * buffer_5_QTMBs = h->buffer_5_QTMBs;

	DIF_segment_desc segdesc;
	segdesc.sequence_number = sequence_number;
	segdesc.block_number = block_number;
	segdesc.ch = ch;
	segdesc.is_MB211 = is_MB211;

	uint8_t * const quantization_number = segdesc.quantization_number;
	uint8_t * const DCT_mode = segdesc.mode;
	uint8_t * const class_number  = segdesc.class_number;

	int dct248_threshold_x16 = h->config.encoder_dct248_threshold*16;
	int dct248_counts = h->config.encoder_dct248_counts;
	int length_test, decrease_QNO;
	int QNO;
	int16_t *core;

	int i;
	int class_numbers_adjusted;

	// do DCT-based tranforming on all 5 MBs and do classification
	for (i = 0; i < (5*6); i++)
	{
		int max_abs_value;
		int block_number = i % 6;

		if (is_MB211 && (block_number == 1 || block_number == 3))
		{
			DCT_mode[i] = 0;
			class_number[i] = 0;
			continue;
		}

		int16_t const * block_src = src_5_MBs + i*64;
		int16_t * block_dst = buffer_5_QTMBs + i*64;

		if (is_interlaced(block_src, dct248_threshold_x16) < dct248_counts)
		{
			dct_88_weighted(block_src, block_dst);
			DCT_mode[i] = 0;
		}
		else
		{
			dct_248_weighted(block_src, block_dst);
			DCT_mode[i] = 1;
		}

		max_abs_value = max_abs_AC(block_dst) / 16;

		int const * const pclassthreshold = 
			(block_number  < 4) ? h->config.encoder_classification_Y :
			(block_number == 4) ? h->config.encoder_classification_V :
			h->config.encoder_classification_U;

		class_number[i] =
			max_abs_value >= pclassthreshold[2] ? 3 :
			max_abs_value >= pclassthreshold[1] ? 2 :
			max_abs_value >= pclassthreshold[0] ? 1 : 0;
	}

	core = core_positive_negative;
	//apply user defined coring_c pattern if enabled in settings
	//coring(buffer_10_TMBs, core, 5*6);

	#define LENGTH_TEST_MAX (5*67*8)

	switch (h->config.encoder_quality_mode)
	{
	case QUALITY_MODE_FAST:
		QNO = 12;
		class_numbers_adjusted=0;

		do {
			length_test = 0;
			decrease_QNO = 0;

			for (i = 0; i < (5*6); i++)
			{
				int block_number = i % 6;

				if (is_MB211 && (block_number == 1 || block_number == 3))
				{
					length_test += 4;  // EOB code
				}
				else
				{
					int16_t const * block = buffer_5_QTMBs + i*64;
					int c1c0 = class_number[i];
					int q_code = quantization_code_table_extended[(c1c0<<4)+QNO];
					int mo = DCT_mode[i];

					int16_t const * q_scale = quantization_scale_extended[mo] + q_code*64;

					weighting_clipping(block, buffer_block, q_scale);
					length_test += get_vlc_length_unrolled[mo](buffer_block);
				}

				if (length_test > LENGTH_TEST_MAX) 
				{
					decrease_QNO = 4;
					break;
				}
			}	
			
			QNO = QNO - decrease_QNO;

			if (QNO < 0)
			{
				if ((QNO + decrease_QNO)!=0)
				{
					QNO = 0;
				}
				else if (!class_numbers_adjusted)
				{
					for (i = 0; i < (5*6); i++)
						class_number[i] = 3;

					class_numbers_adjusted = 1;
					QNO = 8;
				}
				else if ((core - core_positive_negative + 128) < sizeof(core_positive_negative) / sizeof(core_positive_negative[0]))
				{
					QNO = 8;
					core = core + 128;
					coring(buffer_5_QTMBs, core, 5*6);
				}
				else	// should never happen!!!
				{
					QNO = -1;	
					decrease_QNO = 0;
				}
			}
			
		} while (decrease_QNO > 0);
		
		break;

	default:	// QUALITY_MODE_NORMAL
		QNO = 15;
		class_numbers_adjusted = 0;

		do {
			length_test = 0;
			decrease_QNO = 0;

			for (i = 0; i < (5*6); i++)
			{
				int block_number = i % 6;

				if (is_MB211 && (block_number == 1 || block_number == 3))
				{
					length_test += 4;  // EOB code
				}
				else
				{
					int16_t const * block = buffer_5_QTMBs + i*64;
					int c1c0 = class_number[i];
					int q_code = quantization_code_table_extended[(c1c0<<4)+QNO];
					int mo = DCT_mode[i];

					int16_t const * q_scale = quantization_scale_extended[mo] + q_code*64;

					weighting_clipping(block, buffer_block, q_scale);
					length_test += get_vlc_length_unrolled[mo](buffer_block);
				}

				if (length_test > LENGTH_TEST_MAX) 
				{
					decrease_QNO = 1;
					break;
				}
			}	
			
			QNO = QNO - decrease_QNO;

			if (QNO < 0)
			{
				if (!class_numbers_adjusted)
				{
					for (i = 0; i < (5*6); i++)
						class_number[i] = 3;

					class_numbers_adjusted = 1;
					QNO = 15;
				}
				else if ((core - core_positive_negative + 128) < sizeof(core_positive_negative) / sizeof(core_positive_negative[0]))
				{
					QNO = 12;
					core = core + 128;
					coring(buffer_5_QTMBs, core, 5*6);
				}
				else	// should never happen!!!
				{
					QNO = -1;	
					decrease_QNO = 0;
				}
			}
			
		} while (decrease_QNO > 0);
		
		break;
	}

	if (QNO < 0) 
	{
		return;	// for debugging only
	}

	for (i = 0; i < (5*6); i++)
	{
		int block_number = i % 6;

		if (is_MB211 && (block_number == 1 || block_number == 3))
			continue;

		int c1c0 = class_number[i];
		int q_code = quantization_code_table_extended[(c1c0 << 4) + QNO];
		int mo = DCT_mode[i];
		int16_t * block = buffer_5_QTMBs + i * 64;
		int16_t const * q_scale = quantization_scale_extended[mo] + q_code * 64;

		weighting_clipping(block, block, q_scale);
	}		

	for (i = 0; i < 5; i++)
		quantization_number[i] = (uint8_t)QNO;

	encode_videosegment(
		dst_DIF_block,
		buffer_5_QTMBs,
		&segdesc);
}
