/******************************************************************************/
/*                                                                            */
/*  Cedocida DV-Codec                                                         */
/*                                                                            */
/*  Copyright (C) 2005 Andreas Dittrich <dittrich@eit.uni-kl.de>              */
/*  Project start: 15.12.2003                                                 */
/*  Last change  : 15.06.2007                                                 */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/*  This program is free software; you can redistribute it and/or modify      */
/*  it under the terms of the GNU General Public License as published by      */
/*  the Free Software Foundation; either version 2 of the License, or         */
/*  (at your option) any later version.                                       */
/*                                                                            */
/*  This program is distributed in the hope that it will be useful,           */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             */
/*  GNU General Public License for more details.                              */
/*                                                                            */
/*  You should have received a copy of the GNU General Public License         */
/*  along with this program; if not, write to the Free Software               */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
/*                                                                            */
/******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#if defined(_MSC_VER)
#include <windows.h>
#endif

#include "inttypes.h"
#include "config.h"
#include "encode.h"
#include "decode.h"



#define MAX_FILENAME 1024

#define NUM_BYTES_FRAME_DVSD_PAL  (12*150*80)
#define NUM_BYTES_FRAME_DVSD_NTSC (10*150*80)
#define WIDTH_DV 720
#define HEIGHT_PAL 576


#define MIN(a,b) ((a)<(b) ? a : b);


typedef enum {
	mode_create,
	mode_verify
} testmode_t;


typedef struct {

	FILE *hinpfile;
	FILE *houtfile;

	void *henc;
	void *hdec;

	testmode_t testmode;

	int is_PAL;
	int num_bytes_frame_dv;
	int width;
	int height;
	dvformat_t dvfmt;

	uint8_t *pbufdvsd;

} main_context_t;



static void exit_function(void);

#if defined(_MSC_VER)
static BOOL ctrl_handler(DWORD fdwCtrlType);
#endif

static void close_all(main_context_t * const h);
static int test_decoder(main_context_t * const h, char *basename);
static int test_encoder(main_context_t * const h, char *inpfilename, int color_format);

static void open_output_file(main_context_t * const h, char *outfilename);
static int  read_input_data(main_context_t * const h, uint8_t *pdata, int num_bytes_to_read);
static int  write_output_data(main_context_t * const h, uint8_t *pdata, int num_bytes_to_write);

#if !defined(_MSC_VER)
	#define strcpy_s(a,b,c) strcpy(a,c)
	#define strcat_s(a,b,c) strcat(a,c)
#endif



void * g_hmain = NULL;
volatile int g_quit = 0;


int main(int argc, char* argv[])
{
	main_context_t * const h = (main_context_t *)malloc(sizeof(main_context_t));
	if (h == NULL)
	{
		fprintf(stderr, "error: memory allocation failed\n");
		return -4;
	}

	memset(h, 0x00, sizeof(main_context_t));
	g_hmain = (void*)h;

	// init the console handling and set error mode
	int res = atexit(exit_function);
	if (res != 0)
		return -1;

	#if defined(_MSC_VER)

	BOOL bResult = SetConsoleCtrlHandler(
		(PHANDLER_ROUTINE)ctrl_handler,
		TRUE );

	if (!bResult)
		return -1;

	_set_error_mode(_OUT_TO_MSGBOX);

	#endif

	// check arguments 
	///////////////////

	if (argc < 3)
	{
		fprintf(stderr, "usage: testall <-create | -verify> <source_1.dvsd> <source_2.dvsd> ...\n");
		return 0;
	}

	if (strncmp(argv[1], "-create", 7) == 0)
	{
		h->testmode = mode_create;
	}
	else if (strncmp(argv[1], "-verify", 7) == 0)
	{
		h->testmode = mode_verify;
	}
	else
	{
		fprintf(stderr, "error: unknwon option\n");
		return -2;
	}

	int test_failed = 0;

	for (int i = 2; i < argc; i++)
	{
		char * filename = argv[i];

		//open encoder and decoder
		////////////////////////////

		h->henc = encoder_open();
		if (!h->henc)
		{
			fprintf(stderr, "error: opening encoder\n");
			return -4;
		}

		h->hdec = decoder_open();
		if (!h->hdec)
		{
			fprintf(stderr, "error: opening decoder\n");
			return -4;
		}

		//open file to check if its pal or ntsc
		////////////////////////////////////////

		h->hinpfile = fopen(filename, "rb");
		if (!h->hinpfile)
		{
			fprintf(stderr, "error: opening %s\n", filename);
			exit(-1);
		}

		uint8_t header[512];
		int n = (int)fread(header, 1, sizeof(header), h->hinpfile);
		if (n != sizeof(header))
		{
			fprintf(stderr, "error: invalid file %s\n", filename);
			exit(-1);
		}

		uint8_t APT  = header[4] & 0x07;
		uint8_t vsys = header[3] >> 7;
		uint8_t stype = header[5*80 + 3 + 9*5 + 3] & 0x1F;

		h->dvfmt = 
			(APT == 1 && stype == 0x04) ? DVFORMAT_DV50 :
			(APT == 1 && stype == 0x00) ? DVFORMAT_DV25 :
			DVFORMAT_DVSD;

		int nchannels = (h->dvfmt == DVFORMAT_DV50) ? 2 : 1;

		if (vsys == 1)
		{
			h->is_PAL = 1;
			h->num_bytes_frame_dv = 12 * 150 * 80 * nchannels;
			h->width = 720;
			h->height = 576;
		}
		else
		{
			h->is_PAL = 0;
			h->num_bytes_frame_dv = 10 * 150 * 80 * nchannels;
			h->width = 720;
			h->height = 480;
		}

		fclose(h->hinpfile);
		h->hinpfile = NULL;

		//create buffer
		////////////////

		h->pbufdvsd = (uint8_t*)malloc(h->num_bytes_frame_dv);
		if (!h->pbufdvsd)
		{
			fprintf(stderr, "error: memory allocation failed\n");
			return -4;
		}

		//test all
		////////////

		int res = test_decoder(h, filename);

		if (res != 0)
			test_failed = 1;

		close_all(h);
	}

	fprintf(stderr, "\n\n");
	fprintf(stderr, "****************************************\n");
	
	fprintf(stderr, test_failed ? 
		"Test failed!\n" :
		"Test passed!\n");

	fprintf(stderr, "****************************************\n");

	return (test_failed == 0) ? 0 : -1;
}


static void close_all(main_context_t * const h)
{
	if (h->hdec)
	{
		decoder_close(h->hdec);
		h->hdec = NULL;
	}

	if (h->henc)
	{
		encoder_close(h->henc);
		h->henc = NULL;
	}

	if (h->hinpfile)
	{
		fclose(h->hinpfile);
		h->hinpfile = NULL;
	}

	if (h->houtfile)
	{
		fclose(h->houtfile);
		h->houtfile = NULL;
	}

	if (h->pbufdvsd)
	{
		free(h->pbufdvsd);
		h->pbufdvsd = NULL;
	}
}


static void exit_function(void)
{
	main_context_t * const h = (main_context_t *)g_hmain;

	if (!h)
		return;

	close_all(h);
	free(h);
}


#if defined(_MSC_VER)

static BOOL ctrl_handler( DWORD fdwCtrlType ) 
{ 
	switch( fdwCtrlType ) 
	{ 
	case CTRL_C_EVENT: 
	case CTRL_CLOSE_EVENT: 
	case CTRL_BREAK_EVENT:
		g_quit = 1;

		return( TRUE ); 

	case CTRL_LOGOFF_EVENT: 
	case CTRL_SHUTDOWN_EVENT:
	default:
		return FALSE; 
	} 
} 

#endif


static const configvars_t myconfig_default = {
	QUALITY_MODE_NORMAL,	// encoder_quality_mode
	COLORFORMAT_YUY2 | COLORFORMAT_YV12 | COLORFORMAT_RGB32 | COLORFORMAT_RGB24,	// encoder_input_format;
	0,	// encoder_RGB_to_YUV_conversion = 0;
	YV12_CHROMASAMPLING_MPEG2_I,	// encoder_yv12_chroma_sampling;
	25,	// encoder_dct248_threshold = 10;
	15,	// encoder_dct248_counts = 15;
	0,	// aspect_ratio
	0,	// encoder_coring_enable = 0;
	{	// encoder_coring_matrix[64]
	    0,    0,    0,    0,    0,    0,    0,    0,
	    0,    0,    0,    0,    0,    0,    0,    0,
	    0,    0,    0,    0,    0,    0,    0, 1023,
	    0,    0,    0,    0,    0,    0, 1023, 1023,
	    0,    0,    0,    0,    0, 1023, 1023, 1023,
	    0,    0,    0,    0, 1023, 1023, 1023, 1023,
	    0,    0,    0, 1023, 1023, 1023, 1023, 1023,
	    0,    0, 1023, 1023, 1023, 1023, 1023, 1023
	},
	{  0, 10, 210 },	// encoder_classification_Y[3]
	{  0,  0, 3 },	// encoder_classification_V[3]
	{  0,  0, 3 },	// encoder_classification_U[3]
	COLORFORMAT_YUY2 | COLORFORMAT_YV12 | COLORFORMAT_RGB32 | COLORFORMAT_RGB24,	// decoder_output_format
	0,	// decoder_YUV_to_RGB_conversion = 0;
	YV12_CHROMASAMPLING_MPEG2_I	// decoder_YV12_chroma_sampling;
};


static int test_decoder(main_context_t * const h, char *inpfilename)
{
	int iresult = 0;

	struct 
	{
		int color_format;
		int flip_RGB;
		int decoder_YUV_to_RGB_conversion;
		int decoder_YV12_chroma_sampling;
		int bits_per_pixel;
		char const *extension;
	} 
	settings[] =
	{
		{COLORFORMAT_RGB24, 0, RGB2YUV_601, 0, 24,       "__0_235_rgb24"},
		{COLORFORMAT_RGB24, 0, RGB2YUV_UNCHANGED, 0, 24, "__0_255_rgb24"},
		{COLORFORMAT_RGB24, 1, RGB2YUV_601, 0, 24,       "__1_235_rgb24"},
		{COLORFORMAT_RGB24, 1, RGB2YUV_UNCHANGED, 0, 24, "__1_255_rgb24"},
		{COLORFORMAT_RGB32, 0, RGB2YUV_601, 0, 32,       "__0_235_rgb32"},
		{COLORFORMAT_RGB32, 0, RGB2YUV_UNCHANGED, 0, 32, "__0_255_rgb32"},
		{COLORFORMAT_RGB32, 1, RGB2YUV_601, 0, 32,       "__1_235_rgb32"},
		{COLORFORMAT_RGB32, 1, RGB2YUV_UNCHANGED, 0, 32, "__1_255_rgb32"},
		{COLORFORMAT_YUY2, 0, 0, 0, 16, "__yuy2"},
		{COLORFORMAT_YV12, 0, 0, YV12_CHROMASAMPLING_DV, 12,      "__yv12dv"},
		{COLORFORMAT_YV12, 0, 0, YV12_CHROMASAMPLING_MPEG2_P, 12, "__yv12m2p"},
		{COLORFORMAT_YV12, 0, 0, YV12_CHROMASAMPLING_MPEG2_I, 12, "__yv12m2i"}
	};

	int N = sizeof(settings)/sizeof(settings[0]);

	char basename[MAX_FILENAME];
	strcpy_s(basename, MAX_FILENAME, inpfilename);

	char * pch = strrchr(basename, '.');
	if (!pch)
	{
		fprintf(stderr, "error: invalid source filename\n");
		exit(-3);
	}

	pch[0] = 0;

	for (int i = 0; i < N; i++)
	{
		int res = 0;
		configvars_t config = myconfig_default;

		config.decoder_YUV_to_RGB_conversion = settings[i].decoder_YUV_to_RGB_conversion;
		config.decoder_YV12_chroma_sampling  = settings[i].decoder_YV12_chroma_sampling;
		
		decoder_configure(
			h->hdec,
			settings[i].color_format,
			settings[i].flip_RGB,
			h->is_PAL,
			h->dvfmt,
			&config);

		h->hinpfile = fopen(inpfilename, "rb");
		if (!h->hinpfile)
		{
			fprintf(stderr, "error: opening %s\n", inpfilename);
			exit(-1);
		}

		char outfilename[MAX_FILENAME];
		strcpy_s(outfilename, MAX_FILENAME, basename);
		strcat_s(outfilename, MAX_FILENAME, settings[i].extension);
		strcat_s(outfilename, MAX_FILENAME, ".raw");

		open_output_file(h, outfilename);

		int num_bytes_frame_raw = (h->width * h->height * settings[i].bits_per_pixel) / 8;

		uint8_t * pbufraw = (uint8_t*)malloc(num_bytes_frame_raw);
		if (pbufraw == NULL)
			exit(-1);

		for (;;)
		{
			res = read_input_data(h, h->pbufdvsd, h->num_bytes_frame_dv);
			if (res != 0)
				break;

			unsigned int nbytesraw = num_bytes_frame_raw;

			decode_DV(
				h->hdec,
				h->pbufdvsd,
				h->num_bytes_frame_dv,
				pbufraw,
				&nbytesraw);

			res = write_output_data(h, pbufraw, num_bytes_frame_raw);
			if (res != 0)
			{
				iresult = 1;
				break;
			}

			if (g_quit)
				break;
		}

		if (pbufraw)
			free(pbufraw);

		fclose(h->hinpfile);
		h->hinpfile = NULL;

		fclose(h->houtfile);
		h->houtfile = NULL;

		if (g_quit)
			exit(0);

		test_encoder(h, outfilename, settings[i].color_format);
	}

	return iresult;
}


static int test_encoder(main_context_t * const h, char *inpfilename, int color_format)
{
	int iresult = 0;

	struct 
	{
		int color_format;
		int flip_RGB;
		int encoder_quality_mode;
		int encoder_RGB_to_YUV_conversion;
		int encoder_YV12_chroma_sampling;
		int bits_per_pixel;
		char const *extension;
	} 
	settings[] =
	{
		{COLORFORMAT_RGB24, 0, QUALITY_MODE_NORMAL, RGB2YUV_601, 0, 24,       "__n_0_235"},
		{COLORFORMAT_RGB24, 0, QUALITY_MODE_NORMAL, RGB2YUV_UNCHANGED, 0, 24, "__n_0_255"},
		{COLORFORMAT_RGB24, 1, QUALITY_MODE_NORMAL, RGB2YUV_601, 0, 24,       "__n_1_235"},
		{COLORFORMAT_RGB24, 1, QUALITY_MODE_NORMAL, RGB2YUV_UNCHANGED, 0, 24, "__n_1_255"},
		{COLORFORMAT_RGB32, 0, QUALITY_MODE_NORMAL, RGB2YUV_601, 0, 32,       "__n_0_235"},
		{COLORFORMAT_RGB32, 0, QUALITY_MODE_NORMAL, RGB2YUV_UNCHANGED, 0, 32, "__n_0_255"},
		{COLORFORMAT_RGB32, 1, QUALITY_MODE_NORMAL, RGB2YUV_601, 0, 32,       "__n_1_235"},
		{COLORFORMAT_RGB32, 1, QUALITY_MODE_NORMAL, RGB2YUV_UNCHANGED, 0, 32, "__n_1_255"},
		{COLORFORMAT_YUY2, 0, QUALITY_MODE_NORMAL, 0, 0, 16, "__n_yuy2"},
		{COLORFORMAT_YV12, 0, QUALITY_MODE_NORMAL, 0, YV12_CHROMASAMPLING_DV, 12,      "__n_dv"},
		{COLORFORMAT_YV12, 0, QUALITY_MODE_NORMAL, 0, YV12_CHROMASAMPLING_MPEG2_P, 12, "__n_m2p"},
		{COLORFORMAT_YV12, 0, QUALITY_MODE_NORMAL, 0, YV12_CHROMASAMPLING_MPEG2_I, 12, "__n_m2i"},

		{COLORFORMAT_RGB24, 0, QUALITY_MODE_FAST, RGB2YUV_601, 0, 24,       "__f_0_235"},
		{COLORFORMAT_RGB24, 0, QUALITY_MODE_FAST, RGB2YUV_UNCHANGED, 0, 24, "__f_0_255"},
		{COLORFORMAT_RGB24, 1, QUALITY_MODE_FAST, RGB2YUV_601, 0, 24,       "__f_1_235"},
		{COLORFORMAT_RGB24, 1, QUALITY_MODE_FAST, RGB2YUV_UNCHANGED, 0, 24, "__f_1_255"},
		{COLORFORMAT_RGB32, 0, QUALITY_MODE_FAST, RGB2YUV_601, 0, 32,       "__f_0_235"},
		{COLORFORMAT_RGB32, 0, QUALITY_MODE_FAST, RGB2YUV_UNCHANGED, 0, 32, "__f_0_255"},
		{COLORFORMAT_RGB32, 1, QUALITY_MODE_FAST, RGB2YUV_601, 0, 32,       "__f_1_235"},
		{COLORFORMAT_RGB32, 1, QUALITY_MODE_FAST, RGB2YUV_UNCHANGED, 0, 32, "__f_1_255"},
		{COLORFORMAT_YUY2, 0, QUALITY_MODE_FAST, 0, 0, 16, "__f_yuy2"},
		{COLORFORMAT_YV12, 0, QUALITY_MODE_FAST, 0, YV12_CHROMASAMPLING_DV, 12,      "__f_dv"},
		{COLORFORMAT_YV12, 0, QUALITY_MODE_FAST, 0, YV12_CHROMASAMPLING_MPEG2_P, 12, "__f_m2p"},
		{COLORFORMAT_YV12, 0, QUALITY_MODE_FAST, 0, YV12_CHROMASAMPLING_MPEG2_I, 12, "__f_m2i"}
	};

	int N = sizeof(settings)/sizeof(settings[0]);

	char basename[MAX_FILENAME];
	strcpy_s(basename, MAX_FILENAME, inpfilename);

	char * pch = strrchr(basename, '.');
	if (!pch)
	{
		fprintf(stderr, "error: invalid source filename\n");
		exit(-3);
	}

	pch[0] = 0;

	for (int i = 0; i < N; i++)
	{
		if (settings[i].color_format != color_format)
			continue;

		int res = 0;
		configvars_t config = myconfig_default;

		config.encoder_quality_mode = settings[i].encoder_RGB_to_YUV_conversion;
		config.encoder_RGB_to_YUV_conversion = settings[i].encoder_RGB_to_YUV_conversion;
		config.encoder_YV12_chroma_sampling  = settings[i].encoder_YV12_chroma_sampling;

		encoder_configure(
			h->henc,
			settings[i].color_format,
			settings[i].flip_RGB,
			h->is_PAL,
			h->dvfmt,
			&config);

		h->hinpfile = fopen(inpfilename, "rb");
		if (!h->hinpfile)
		{
			fprintf(stderr, "error: opening %s\n", inpfilename);
			exit(-1);
		}

		char outfilename[MAX_FILENAME];
		strcpy_s(outfilename, MAX_FILENAME, basename);
		strcat_s(outfilename, MAX_FILENAME, settings[i].extension);
		strcat_s(outfilename, MAX_FILENAME, ".dvsd");

		open_output_file(h, outfilename);

		int num_bytes_frame_raw = (h->width * h->height * settings[i].bits_per_pixel) / 8;

		uint8_t * pbufraw = (uint8_t*)malloc(num_bytes_frame_raw);
		if (pbufraw == NULL)
			exit(-1);

		for (int loopcounter = 0;; loopcounter++)
		{
			res = read_input_data(h, pbufraw, num_bytes_frame_raw);
			if (res != 0)
				break;

			unsigned int nbytesdv = h->num_bytes_frame_dv;

			encode_DV(
				h->henc,
				pbufraw,
				num_bytes_frame_raw,
				h->pbufdvsd,
				&nbytesdv);

			res = write_output_data(h, h->pbufdvsd, h->num_bytes_frame_dv);
			if (res != 0)
			{
				iresult = 1;
				break;
			}

			if (g_quit)
				break;
		}

		if (pbufraw)
			free(pbufraw);

		fclose(h->hinpfile);
		h->hinpfile = NULL;

		fclose(h->houtfile);
		h->houtfile = NULL;

		if (g_quit)
			exit(0);
	}

	return iresult;
}


static int read_input_data(main_context_t * const h, uint8_t *pdata, int num_bytes_to_read)
{
	int num_bytes_read = (int)fread(pdata, 1, num_bytes_to_read, h->hinpfile);
	if (num_bytes_read == 0)	//eof
		return 1;

	if (num_bytes_read != num_bytes_to_read)
	{
		fprintf(stderr, "error: reading file\n");
		exit(-1);
	}

	return 0;
}


static void open_output_file(main_context_t * const h, char *outfilename)
{
	char const * mode = NULL; 
	
	if (h->testmode == mode_create)
	{
		printf("creating file %s\n", outfilename);
		mode = "wb";
	}
	else
	{
		printf("verifying file %s\n", outfilename);
		mode = "rb";
	}

	h->houtfile = fopen(outfilename, mode);
	if (!h->houtfile)
	{
		fprintf(stderr, "error: opening %s\n", outfilename);
		exit(-1);
	}
}

static int write_output_data(main_context_t * const h, uint8_t *pdata, int num_bytes_to_write)
{
	if (h->testmode == mode_create)
	{
		int num_bytes_written = (int)fwrite(pdata, 1, num_bytes_to_write, h->houtfile);
		if (num_bytes_written != num_bytes_to_write)
		{
			fprintf(stderr, "error: writing file\n");
			exit(-1);
		}
	}
	else
	{
		uint32_t chunk[256];
		uint8_t * pbuf = (uint8_t*)chunk;

		for (;;)
		{
			int num_bytes_to_read = MIN((int)sizeof(chunk), num_bytes_to_write);
			if (num_bytes_to_read <= 0)
				break;

			int num_bytes_read = (int)fread(pbuf, 1, num_bytes_to_read, h->houtfile);
			if (num_bytes_read != num_bytes_to_read)
			{
				fprintf(stderr, "warning: filesize missmatch\n");
				return 1;
			}

			if (memcmp(pbuf, pdata, num_bytes_read) != 0)
			{
				fprintf(stderr, "warning: file missmatch\n");
				return 1;
			}

			pdata += num_bytes_read;
			num_bytes_to_write -= num_bytes_read;
		}
	}

	return 0;
}
